local classes = require("lua/explosionsNG/helpers/classes")
local LinearPath = require("lua/explosionsNG/pathing/LinearPath")

-- TODO Unused
--- @class CylindricalArea
--- @field private centre LuaVec3 The centre point of the cylinder
--- @field private radius number The radius of the cylinder
--- @field private INCREMENT_MULTIPLIER number __Static__ The increment step to use when moving a point outside this area into it. A smaller value will increase precision but take more computation.
local CylindricalArea = classes.newClass()
CylindricalArea.INCREMENT_MULTIPLIER = 0.1

--- Returns the distance along the X and Y axis between this cylinder's centre and the given position
--- @param position LuaVec3 The position
--- @return number the distance along the X and Y axis between this cylinder's centre and the given position
function CylindricalArea:xyDistance(position)
  local xyPosition = position:z0()
  return self.centre:distance(xyPosition)
end

--- Returns whether this cylinder area contains the given position or not
--- @param position LuaVec3 The position to check
--- @return boolean Whether this cylinder contains the given position or not
function CylindricalArea:contains(position)
  return self:xyDistance(position) <= self.radius
end

--- Moves a point to inside this cylinder's area incrementally and returns the new point. The precision of the increment
--- is determined by the `INCREMENT_MULTIPLIER`.
--- @param position LuaVec3 The position to move inside
--- @return LuaVec3 The original position, moved to inside this cylinder
--- @private
function CylindricalArea:moveUntilInside(position)
  local increment = self:xyDistance(position) * CylindricalArea.INCREMENT_MULTIPLIER
  local linearPath = LinearPath:new(position, self.centre, increment)

  while not self:contains(linearPath:getCurrentPosition()) do
    linearPath:addDistance(increment)
  end

  return linearPath:getCurrentPosition()
end

--- Returns the closest point inside this cylinder to the given position. If this area already contains the position,
--- then it will be returned unchanged.
--- @param position LuaVec3 The position to get the closest point in this cylinder to
--- @return LuaVec3 The closest point inside this cylinder to the position
function CylindricalArea:findClosest(position)
  if self:contains(position) then return position end
  return self:moveUntilInside(position)
end

--- @return LuaVec3 The centre point of the cylinder area
function CylindricalArea:getCentre()
  return vec3(self.centre)
end

--- __Constructor__
--- @param centre LuaVec3 The centre point of the cylinder
--- @param radius number The radius of the cylinder
--- @return CylindricalArea The constructed instance
function CylindricalArea:new(centre, radius)
  local self = classes.setmetatable({}, CylindricalArea)
  self.centre = vec3(centre):z0()
  self.radius = radius
  return self
end

return CylindricalArea
