local classes = require("lua/explosionsNG/helpers/classes")
local ExplosiveStats = require("lua/explosionsNG/explosives/base/ExplosiveStats")
local Anomaly = require("lua/explosionsNG/worldObjects/anomalies/Anomaly")
local AttackingAnomaly = require("lua/explosionsNG/worldObjects/anomalies/attacks/AttackingAnomaly")
local AnomalyExplosive = require("lua/explosionsNG/explosives/contained/anomaly/AnomalyExplosive")
local ShockwaveExplosiveStats = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosiveStats")

--- It should be noted that this class does NOT extend the `SuperExplosiveStats` class as, although The Anomaly could be
--- considered a super-explosive, its stat's do not behave the same way
--- @class AnomalyExplosiveStats : ExplosiveStats<AnomalyOptions>
local AnomalyExplosiveStats = classes.newClass(ExplosiveStats)

--- @return number The absolute energy stat
--- @see ExplosiveStats#getAbsoluteEnergyStat
function AnomalyExplosiveStats:getAbsoluteEnergyStat()
  --- @type ShockwaveExplosiveOptions
  local dischargeShockwaveOptions = {
    time = AnomalyExplosive.DISCHARGE_TIME,
    forceMultiplier = AttackingAnomaly.ATTACK_FORCE_MULTIPLIER,
    maxRadius = AnomalyExplosive:getDischargeRange(self.explosiveOptions.scale)
  }
  local averageNumberOfDischarges = AnomalyExplosive:getMaximumNumberOfDischarges(self.explosiveOptions.life)
  local dischargeShockwave = ShockwaveExplosiveStats:new(dischargeShockwaveOptions)
  return dischargeShockwave:getAbsoluteEnergyStat() * averageNumberOfDischarges
end

--- @return number The absolute range stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function AnomalyExplosiveStats:getAbsoluteRangeStat()
  local scale = self.explosiveOptions.scale
  local dischargeRange = AnomalyExplosive:getDischargeRange(scale)
  local empRange = AnomalyExplosive:getEmpRange(scale)
  return math.max(dischargeRange, empRange)
end

--- @return number The absolute accuracy stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function AnomalyExplosiveStats:getAbsoluteAccuracyStat()
  local shockwaveRange = AnomalyExplosive:getDischargeRange(self.explosiveOptions.scale)

  local shockwaveDelay = Anomaly.DISCHARGING_PERIOD / 2
  local totalTime = shockwaveDelay + AnomalyExplosive.DISCHARGE_TIME

  -- avg.speed = d/t
  return shockwaveRange / totalTime
end

--- @return number The absolute attack rate stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function AnomalyExplosiveStats:getAbsoluteAttackRateStat()
  return Anomaly.DISCHARGING_PERIOD
end

--- __Constructor__
--- @param explosiveOptions AnomalyOptions The explosive options to get the stats of
--- @return AnomalyExplosiveStats The constructed instance
function AnomalyExplosiveStats:new(explosiveOptions)
  local self = classes.setmetatable(ExplosiveStats:new(explosiveOptions, "averageSpeed", "multiple"), AnomalyExplosiveStats)
  return self
end

return AnomalyExplosiveStats
