local classes = require("lua/explosionsNG/helpers/classes")
local SuperExplosiveStats = require("lua/explosionsNG/explosives/base/SuperExplosiveStats")
local MineExplosiveStats = require("lua/explosionsNG/explosives/composed/mine/MineExplosiveStats")
local ProjectileVolcano = require("lua/explosionsNG/projectiles/groups/ProjectileVolcano")

--- @class LandMineExplosiveStats : SuperExplosiveStats<LandMineExplosiveOptions, ShockwaveExplosiveOptions, MineExplosiveStats>
--- @field mineStats MineExplosiveStats The stats for the mines that the land mine explosive produces
local LandMineExplosiveStats = classes.newClass(SuperExplosiveStats)

--- @return number The number of sub-explosives that this super explosive creates
--- @see SuperExplosiveStats#getNumberOfSubExplosives
function LandMineExplosiveStats:getNumberOfSubExplosives()
  return self.explosiveOptions.numberOfMines
end

--- @return number The maximum range of the super-explosive itself. In general, this should be how far a sub-explosive can be placed by the super-explosive
--- @see SuperExplosiveStats#getSuperExplosiveRange
function LandMineExplosiveStats:getSuperExplosiveRange()
  return SuperExplosiveStats:getMaximumRangeOfProjectileVolcano(self.explosiveOptions.ejectionMultiplier)
end

--- @return number The absolute attack rate stat
--- @see Explosive#getAbsoluteAttackRateStat
function LandMineExplosiveStats:getAbsoluteAttackRateStat()
  local lifeTime = ProjectileVolcano:getTotalProjectileAirTime()
  return lifeTime / self.explosiveOptions.numberOfMines
end

--- __Constructor__
--- @param explosiveOptions LandMineExplosiveOptions The explosive options to get the stats of
--- @return LandMineExplosiveStats The constructed instance
function LandMineExplosiveStats:new(explosiveOptions)
  local mineStats = MineExplosiveStats:new(explosiveOptions.mineOptions)
  local self = classes.setmetatable(SuperExplosiveStats:new(explosiveOptions, mineStats), LandMineExplosiveStats)
  return self
end

return LandMineExplosiveStats
