local classes = require("lua/explosionsNG/helpers/classes")
local SuperExplosiveStats = require("lua/explosionsNG/explosives/base/SuperExplosiveStats")
local LightningExplosive = require("lua/explosionsNG/explosives/composed/lightning/LightningExplosive")
local LightningExplosiveStats = require("lua/explosionsNG/explosives/composed/lightning/LightningExplosiveStats")
local LightningStormExplosive = require("lua/explosionsNG/explosives/contained/lightningStorm/LightningStormExplosive")

--- @class LightningStormExplosiveStats : SuperExplosiveStats<LightningStormExplosiveOptions, ShockwaveExplosiveOptions, LightningStormExplosiveStats>
--- @field lightningStats LightningExplosiveStats The stats for an individual lighting bolt
local LightningStormExplosiveStats = classes.newClass(SuperExplosiveStats)

--- @return number The number of sub-explosives that this super explosive creates
--- @see SuperExplosiveStats#getNumberOfSubExplosives
function LightningStormExplosiveStats:getNumberOfSubExplosives()
  return math.floor(self.explosiveOptions.life / self.explosiveOptions.attackDelay)
end

--- @return number The maximum range of the super-explosive itself. In general, this should be how far a sub-explosive can be placed by the super-explosive
--- @see SuperExplosiveStats#getSuperExplosiveRange
function LightningStormExplosiveStats:getSuperExplosiveRange()
  return self.explosiveOptions.radius
end

--- @return number The absolute attack rate stat
--- @see ExplosiveStats#getAbsoluteAttackRateStat
function LightningStormExplosiveStats:getAbsoluteAttackRateStat()
  return self.explosiveOptions.attackDelay
end

--- __Constructor__
--- @param explosiveOptions LightningStormExplosiveOptions The explosive options to get the stats of
--- @return LightningStormExplosiveStats The constructed instance
function LightningStormExplosiveStats:new(explosiveOptions)
  local lightningPreset = LightningExplosive:getPresetOptions(LightningStormExplosive.LIGHTING_PRESET)
  local lightningStats = LightningExplosiveStats:new(lightningPreset)
  local self = classes.setmetatable(SuperExplosiveStats:new(explosiveOptions, lightningStats), LightningStormExplosiveStats)
  return self
end

return LightningStormExplosiveStats
