local classes = require("lua/explosionsNG/helpers/classes")
local ExplosiveStats = require("lua/explosionsNG/explosives/base/ExplosiveStats")
local ShockwaveExplosiveStats = require("lua/explosionsNG/explosives/contained/shockwave/ShockwaveExplosiveStats")

--- @class OrbitalStrikeExplosiveStats : ExplosiveStats<OrbitalStrikeExplosiveOptions>
--- @field private shockwaveStats ShockwaveExplosiveStats The stats for the individual shockwaves
local OrbitalStrikeExplosiveStats = classes.newClass(ExplosiveStats)

--- @return number The absolute energy stat
--- @see ExplosiveStats#getAbsoluteEnergyStat
function OrbitalStrikeExplosiveStats:getAbsoluteEnergyStat()
  return self.shockwaveStats:getAbsoluteEnergyStat() * self.explosiveOptions.pulsarOptions.numberOfShockwaves
end

--- @return number The absolute range stat
--- @see ExplosiveStats#getAbsoluteRangeStat
function OrbitalStrikeExplosiveStats:getAbsoluteRangeStat()
  return self.explosiveOptions.pulsarOptions.maxRadius
end

--- @return number The absolute accuracy stat
--- @see ExplosiveStats#getAbsoluteAccuracyStat
function OrbitalStrikeExplosiveStats:getAbsoluteAccuracyStat()
  return self.shockwaveStats:getAbsoluteAccuracyStat()
end

--- @return number The absolute attack rate stat
--- @see ExplosiveStats#getAbsoluteAttackRateStat
function OrbitalStrikeExplosiveStats:getAbsoluteAttackRateStat()
  return self.explosiveOptions.pulsarOptions.interval
end

--- __Constructor__
--- @param explosiveOptions OrbitalStrikeExplosiveOptions The explosive options to get the stats of
--- @return OrbitalStrikeExplosiveStats The constructed instance
function OrbitalStrikeExplosiveStats:new(explosiveOptions)
  local self = classes.setmetatable(ExplosiveStats:new(explosiveOptions, "averageSpeed", "multiple"), OrbitalStrikeExplosiveStats)
  self.shockwaveStats = ShockwaveExplosiveStats:new({
    forceMultiplier = explosiveOptions.forceMultiplier,
    maxRadius = explosiveOptions.pulsarOptions.maxRadius,
    time = explosiveOptions.pulsarOptions.time
  })
  return self
end

return OrbitalStrikeExplosiveStats
