local classes = require("lua/explosionsNG/helpers/classes")
local effectFactory = require("lua/explosionsNG/vehicles/effects/factory")
local ExplosiveStats = require("lua/explosionsNG/explosives/base/ExplosiveStats")

--- @class ShockwaveExplosiveStats : ExplosiveStats<ShockwaveExplosiveOptions>
local ShockwaveExplosiveStats = classes.newClass(ExplosiveStats)

--- @return number The absolute energy stat
--- @see Explosive#getEnergyStat
function ShockwaveExplosiveStats:getAbsoluteEnergyStat()
  return effectFactory.BASE_FORCE * self.explosiveOptions.forceMultiplier
end

--- @return number The absolute range stat
--- @see Explosive#getRangeStat
function ShockwaveExplosiveStats:getAbsoluteRangeStat()
  return self.explosiveOptions.maxRadius
end

--- @return number The absolute accuracy stat
--- @see Explosive#getAccuracyStat
function ShockwaveExplosiveStats:getAbsoluteAccuracyStat()
  local shockwaveOptions = self.explosiveOptions.shockwaveOptions

  if shockwaveOptions and (--[[--- @not nil]] shockwaveOptions).implode then
    return math.huge -- Impossible to avoid an implosion if a vehicle is assumed to start inside of it
  end

  return self.explosiveOptions.maxRadius / self.explosiveOptions.time
end

--- @return number The absolute attack rate stat
--- @see Explosive#getAbsoluteAttackRateStat
function ShockwaveExplosiveStats:getAbsoluteAttackRateStat()
  return ExplosiveStats.SINGLE_ATTACK_RATE_VALUE
end

--- __Constructor__
--- @param explosiveOptions ShockwaveExplosiveOptions The explosive options to get the stats of
--- @return ShockwaveExplosiveStats The constructed instance
function ShockwaveExplosiveStats:new(explosiveOptions)
  local self = classes.setmetatable(ExplosiveStats:new(explosiveOptions), ShockwaveExplosiveStats)
  return self
end

return ShockwaveExplosiveStats
