local Array = require("lua/explosionsNG/misc/Array")

--- @alias SphericalBound { offset: LuaVec3, radius: number }
--- @alias BoundingArea number | SphericalBound[] @A bounding area can either be a single sphere with a uniform radius, or an array of spheres with offsets

--- Returns the offset that should be added to the elevation of an object with the given bounds
--- @param bounds BoundingArea | nil The bounds
--- @return LuaVec3 The elevation offset of the bounds
local function getElevationOffset(bounds)
  if not bounds then return vec3() end
  if type(bounds) == "number" then return vec3(0, 0, --[[--- @type number ]] bounds) end

  local lowestBound = Array:new(--[[--- @type SphericalBound[] ]] bounds):findByComparison(function(a, b)
    return a.offset.z - a.radius < b.offset.z - b.radius and a or b
  end)
  -- The lowest z value will be the offset value plus the radius of the sphere itself to get the bottommost point on the sphere
  return vec3(0, 0, lowestBound.offset.z + lowestBound.radius)
end

return {
  getElevationOffset = getElevationOffset
}