--- Constructs a new class and mimics inheritance for Lua 'classes'.
--- Note that there is no return type documentation because type-checkers get confused with inheritance for any return
--- other than `any` :/
--- @param subclass? any Optional class that this class will inherit from
local function createClass(subclass)
	local Class = {}
	Class.__index = Class

	if subclass then
		setmetatable(Class, { __index = subclass })
	end

	return Class
end

--- Works the sames as the in-built `setmetatable`, however correctly types the resultant table to prevent the need for
--- type-casting everywhere. However, this does not work for generic tables, i.e. when generic type `T` is a type that
--- contains other generics
--- @generic T, S
--- @param table T The table to set the metatable of
--- @param metatable S The table to set as the metatable
--- @return S The returned table
local function typedSetmetatable(table, metatable)


	return --[[--- @type S ]] setmetatable(table, --[[--- @type table ]] metatable)
end

return {
	newClass = createClass,
	setmetatable = typedSetmetatable
}
