local strings = require("lua/explosionsNG/helpers/strings")

-- The relative path to the materials file that all levels should adhere too
local MATERIAL_RELATIVE_PATH = "/explosionsNG/main.materials.json"

--- Checks and returns whether the given level has materials generated/set up for use with this mod
--- @param levelPath string The level path to check
local function areMaterialGeneratedForLevel(levelPath)
  return FS:fileExists(levelPath..MATERIAL_RELATIVE_PATH)
end

--- Creates a materials file for the level with the given path.
--- @param levelPath string The level path to attempt to create materials for
local function createMaterialsForLevel(levelPath)
  local EXPECTED_PREFIX = "/levels"
  -- smallgrid comes with this mod so we can copy from it
  local COPY_FROM = "levels/smallgrid"..MATERIAL_RELATIVE_PATH

  if not strings.startsWith(levelPath, EXPECTED_PREFIX) then
    log("W", "explosivesNG", "This level does not start with the expected prefix. This is either a bug in BeamNG or is a modded map causing issues. level="..tostring(levelPath))
    return
  end

  local copyTo = levelPath..MATERIAL_RELATIVE_PATH
  log("I", "explosivesNG", "Generating materials for "..levelPath)
  FS:copyFile(COPY_FROM, copyTo)
end

--- Checks and returns whether the current level has materials set up for this mod to load models correctly. If it does
--- not, this function will create the materials. In this case, this function will __still__ return `false` to indicate
--- that the materials need loading again.
--- @return boolean Whether materials were set up for the current level, before this function was called
local function checkAndCopyMaterialForCurrent()
  local levelPath = getMissionFilename():gsub("info.json", "")
  local areMaterialsSetUp = areMaterialGeneratedForLevel(levelPath)

  if not areMaterialsSetUp then
    createMaterialsForLevel(levelPath)
  end

  return areMaterialsSetUp
end

return {
  createMaterialsForLevel = createMaterialsForLevel,
  checkAndCopyMaterialForCurrentLevel = checkAndCopyMaterialForCurrent
}