--- Splits a string into an array of substrings based on the given separator
--- @param str string The string to split
--- @param separator? string The separator. Defaults to `" "`
--- @return string[] The split string as substrings
local function split(str, separator)
  separator = separator or " "
  local fields = --[[--- @type string[] ]] {}

  local pattern = string.format("([^%s]+)", separator)
  --- @type fun: nil
  local repl = function(c) fields[#fields + 1] = c end
  string.gsub(str, pattern, repl)

  return fields
end

--- Returns whether the given string begins the given prefix
--- @param string string The string to search
--- @param prefix string The prefix to check for
--- @return boolean Whether the string starts with the prefix
local function startsWith(string, prefix)
  return string:sub(1, #prefix) == prefix;
end

return {
  split = split,
  startsWith = startsWith
}