local classes = require("lua/explosionsNG/helpers/classes")

--- @class Events<T> : Registrable<T>
--- @field private eventHandlers { [T]: (fun(context: any): void) | nil } A map of event to handler to call
local Events = classes.newClass()

--- Registers an event handler to one of the given events
--- @param event T The event to register a handler to
--- @param eventHandler fun(context: any): void The handler to set
function Events:registerEventHandler(event, eventHandler)
  if self.eventHandlers[event] then
    log("W", "explosivesNG", "This event already has an event handler for the "..tostring(event).." event")
    return
  end

  self.eventHandlers[event] = eventHandler
end

--- Unregisters an event handler from one of the given events
--- @param event T The event to unregister a handler to
--- @param eventHandler fun(): void The handler to remove
function Events:unregisterEventHandler(event, eventHandler)
  local existingHandler = self.eventHandlers[event]
  if existingHandler and existingHandler == eventHandler then
    self.eventHandlers[event] = nil
    return
  end

  log("W", "explosivesNG", "Cannot unregister the "..tostring(event).." event as it does not match the given event handler")
end

--- Notifies the handler associated with the given event, if it exist
--- @param event T The event to emit
--- @param context? any Any additional context to send with the event
function Events:emit(event, context)
  local handler = self.eventHandlers[event]
  if handler then handler(context) end
end

--- __Constructor__
--- @generic S
--- @return Events<S> The constructed instance
function Events:new()
  local self = --[[--- @type Events<S> ]] classes.setmetatable({}, Events)
  self.eventHandlers = {}
  return self
end

return Events
