local explosivesStore = require("lua/explosionsNG/stores/explosives/store")

local explosivesData = jsonDecode(readFile("lua/explosionsNG/stores/explosives/data/explosives.json"))
local propertiesData = jsonDecode(readFile("lua/explosionsNG/stores/explosives/data/properties.json"))

--- @alias ExplosivePresetJson { id: string, name: string, rank: number }
--- @alias ExplosiveJson { id: string, name: string, tier: number, description: string, properties: string[], presets: ExplosivePresetJson[] }

--- @alias ExplosivePresetDto { id: string, name: string, range: number, energy: number, accuracy: number, attackRate: number, rank: number }
--- @alias ExplosivePropertyDto { id: string, name: string, description: string }
--- @alias ExplosiveDto { id: string, name: string, tier: number, description: string, properties: ExplosivePropertyDto[], presets: ExplosivePresetDto[] }

--- Finds the `ExplosivePropertyDto`, given its id
--- @param propertyId string The id of the property to find
--- @return ExplosivePropertyDto The found explosive property
local function findProperty(propertyId)
  for i = 1, #propertiesData.properties do
    local prop = propertiesData.properties[i]
    if prop.id == propertyId then
      return prop
    end
  end

  log("E", "explosivesNG", "findProperty: could not find property with id="..tostring(propertyId))
  return nil
end

--- Converts an `ExplosivePresetJson` object to an `ExplosivePresetDto` object
--- @param explosiveId string The id of the associated explosive that the preset belongs to
--- @param preset ExplosivePresetJson The preset to convert
--- @return ExplosivePresetDto The converted preset
local function jsonPresetToDto(explosiveId, preset)
  local stats = explosivesStore.getExplosivePresetStats(explosiveId, preset.id)
  --local averageOfStats = (stats.energy + stats.range + stats.accuracy + stats.attackRate) / 4
  --dump(explosiveId, preset.id, preset.rank.." | "..math.max(math.floor(averageOfStats * 10), 1))

  return {
    id = preset.id,
    name = preset.name,
    rank = preset.rank,
    energy = stats.energy,
    range = stats.range,
    accuracy = stats.accuracy,
    attackRate = stats.attackRate
  }
end

--- Converts an `ExplosiveJson` object to an `ExplosiveDto` object
--- @param explosive ExplosiveJson The explosive to convert
--- @return ExplosiveDto The converted explosive
local function jsonToDto(explosive)
  -- Look up properties
  local properties = --[[--- @type ExplosivePropertyDto[] ]] {}
  for i = 1, #explosive.properties do
    local property = explosive.properties[i]
    local dto = findProperty(property)
    table.insert(properties, dto)
  end

  -- Look up preset stats
  local presets = --[[--- @type ExplosivePresetDto[] ]] {}
  for i = 1, #explosive.presets do
    local preset = explosive.presets[i]
    local dto = jsonPresetToDto(explosive.id, preset)
    table.insert(presets, dto)
  end

  return {
    id = explosive.id,
    name = explosive.name,
    tier = explosive.tier,
    description = explosive.description,
    presets = presets,
    properties = properties
  }
end

--- @return ExplosiveDto[] The total list of explosives
local function getListOfExplosives()
  local data = --[[--- @type ExplosiveDto[] ]] {}

  for i = 1, #explosivesData.explosives do
    local explosive = explosivesData.explosives[i]
    local dto = jsonToDto(explosive)
    table.insert(data, dto)
  end

  return data
end

return {
  getListOfExplosives = getListOfExplosives
}