local classes = require("lua/explosionsNG/helpers/classes")
local Timer = require("lua/explosionsNG/timers/Timer")

--- A timer that repeats indefinitely at a fixed interval period
--- @class IntervalTimer : Timer
local IntervalTimer = classes.newClass(Timer)

--- Restarts the timer and calls the `callback`.
function IntervalTimer:onTimerFinished()
  self:restart()
  self.callback()
end

--- __Constructor__
--- @param timeLimit number The time limit for each interval cycle of the timer
--- @param callback fun(): void A callback function to call after the timer finishes each interval cycle
--- @param startImmediately? boolean Whether to start the timer immediately
--- @return IntervalTimer The constructed instance
function IntervalTimer:new(timeLimit, callback, startImmediately)
  local self = classes.setmetatable(Timer:new(timeLimit, callback, startImmediately), IntervalTimer)
  return self
end

return IntervalTimer
