local classes = require("lua/explosionsNG/helpers/classes")
local BoundingBox = require("lua/explosionsNG/misc/BoundingBox")

--- A wrapper to a `BoundingBox` object that is specifically catered towards vehicles
--- @class VehicleBounds
--- @field private boundingBox BoundingBox The vehicle bounding box
local VehicleBounds = classes.newClass()

--- Changes the `boundingBox`'s rotation and position from the vehicle's `position` and `direction`
---@param position LuaVec3 The position to change the bounding box's centre to
---@param direction LuaQuat The direction to change the bounding box to face
function VehicleBounds:updateFromVehicle(position, direction)
  self.boundingBox:changeCentre(position)
  self.boundingBox:changeRotation(direction)
end

--- Calls and returns the result of `isIntersectingSphere` with the `bounds`
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @param radius number The radius of the sphere
function VehicleBounds:isIntersectingSphere(centre, radius)
  return self.boundingBox:isIntersectingSphere(centre, radius)
end

--- Returns the closest point on this bounding box (in the absolute coordinate system) with respect to a sphere's centre
--- position.
--- @param centre LuaVec3 The centre of the sphere. This must be defined in the absolute coordinate system
--- @return LuaVec3 The closest point to the sphere `centre`
function VehicleBounds:findClosestPointToSphere(centre)
  return self.boundingBox:findClosestPointToSphere(centre)
end

--- Calls `display` on the `boundingBox`
function VehicleBounds:display()
  self.boundingBox:debugDisplay()
end

--- __Constructor__
--- @return VehicleBounds The constructed instance
function VehicleBounds:new(nodes, position, direction)
  local self = classes.setmetatable({}, VehicleBounds)
  self.boundingBox = BoundingBox:fromPoints(nodes, position, direction)
  return self
end

return VehicleBounds
