local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")

--- Represents a beamstate planet (name isn't accurate to what it actually is - it is a point mass that exerts
--- gravitational on a vehicle; not strictly a planet).
---
--- This is a wrapper to the object that beamstate wants so that it is typed correctly and much neater.
--- @class VehiclePlanet
--- @field private position LuaVec3 The position of the planet
--- @field private radius number The radius (in m) of the planet
--- @field private mass number The mass (in kg) of the planet
local VehiclePlanet = classes.newClass()

--- __Static__
---
--- Maps and returns a string representation that can be consumed by beamstate, of a list of `VehiclePlanets`
--- @param array Array<VehiclePlanet> The list of planets
--- @return string The string representation
function VehiclePlanet:arrayToBeamstateString(array)
  local beamstatePlanets = array:map(VehiclePlanet.toBeamstate)
  local planetsArray = Array:flatten(beamstatePlanets):toPrimitive()
  return dumps(planetsArray)
end

--- @return Array<LuaVec3 | number> The beamstate API version of this planet
function VehiclePlanet:toBeamstate()
  return Array:new({ vec3(self.position), self.radius, self.mass })
end

--- __Constructor__
--- @param position LuaVec3 The position of the planet
--- @param radius number The radius (in m) of the planet
--- @param mass number The mass (in kg) of the planet
--- @return VehiclePlanet The constructed instance
function VehiclePlanet:new(position, radius, mass)
  local self = classes.setmetatable({}, VehiclePlanet)
  self.position = position
  self.radius = radius
  self.mass = mass
  return self
end

return VehiclePlanet
