local classes = require("lua/explosionsNG/helpers/classes")
local VehicleNodes = require("lua/explosionsNG/vehicles/VehicleNodes")
local VehicleEffect = require("lua/explosionsNG/vehicles/effects/VehicleEffect")

--- @class VehicleParticleEffect : VehicleEffect
--- @field private vehicleNodes VehicleNodes The vehicle's nodes that this effect is working on
--- @field private particleType VehicleParticleType The type of particle to use
local VehicleParticleEffect = classes.newClass(VehicleEffect)

--- Applies particle effects to a random node on the vehicle's nodes.
--- If the vehicle nodes are not yet created, then this method will first create theme.
--- @param vehicle Vehicle The vehicle to affect
--- @private
function VehicleParticleEffect:applyParticles(vehicle)
  if not self.vehicleNodes then
    local vehicleObject = vehicle:getGameObject()
    self.vehicleNodes = VehicleNodes:new(vehicleObject)
  end

  self.vehicleNodes:applyParticlesToRandomNode(self.particleType)
end

--- __Constructor__
--- @param sourceId string The id of the object that is the source of the effect
--- @param time number The amount of time the effect is in action
--- @param particleType VehicleParticleType The type of particle to use
--- @return VehicleParticleEffect The constructed instance
function VehicleParticleEffect:new(sourceId, time, particleType)
  local self = classes.setmetatable(VehicleEffect:new(sourceId, nil, { effectDuration = time }), VehicleParticleEffect)
  -- This is not passed in the super call and set here as `self` is not defined when passing in as a parameter when calling super
  self.affectFn = function(vehicle) self:applyParticles(vehicle) end
  self.particleType = particleType
  self.vehicleNodes = nil
  return self
end

return VehicleParticleEffect
