local classes = require("lua/explosionsNG/helpers/classes")
local VehicleEffect = require("lua/explosionsNG/vehicles/effects/VehicleEffect")

--- @class VehicleSoundEffect : VehicleEffect
local VehicleSoundEffect = classes.newClass(VehicleEffect)

--- Plays a sound through the audio engine. If the given vehicle is not a player vehicle, then no sound is played.
--- @param vehicle Vehicle The vehicle to affect
--- @param track string The track to play
function VehicleSoundEffect:playSoundOnPlayer(vehicle, track)
  if not vehicle:isPlayer() then return end
  Engine.Audio.playOnce("AudioGui", "art/explosionsNG/sounds/"..track..".ogg")
end

--- __Constructor__
--- @param sourceId string The id of the object that is the source of the effect
--- @return VehicleSoundEffect The constructed instance
function VehicleSoundEffect:new(sourceId, track)
  --- @type VehicleAffectFn
  local affectFn = function(vehicle) VehicleSoundEffect:playSoundOnPlayer(vehicle, track) end
  local self = classes.setmetatable(VehicleEffect:newInstantaneousEffect(sourceId, affectFn), VehicleSoundEffect)
  return self
end

return VehicleSoundEffect
