local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local ConvergencePhase = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhase")
local AssaultAttackPattern = require("lua/explosionsNG/worldObjects/anomalies/attacks/AssaultAttackPattern")

--- @class ConvergencePhaseOne : ConvergencePhase
local ConvergencePhaseOne = classes.newClass(ConvergencePhase)

--- __Static__
---
--- Returns the attack type of the given subphase
--- @param subphase number The subphase to get the attack type of
--- @return AttackingAnomalyAttack The attack type
--- @private
function ConvergencePhaseOne:getSubPhaseAttack(subphase)
  if subphase == 1 then return "emp" end
  if subphase == 2 then return "fire" end
  return "dissolve"
end

--- @return number The number of subphases
--- @see ConvergencePhase#getNumberOfSubphases
function ConvergencePhaseOne:getNumberOfSubphases()
  return 3
end

--- @return AttackPattern A new attack pattern for a new phase
--- @see ConvergencePhase#createAttackPattern
function ConvergencePhaseOne:createAttackPattern()
  --[[--- @type AssaultAttackPatternOptions ]] local options = {
    type = ConvergencePhaseOne:getSubPhaseAttack(self.subphase),
    time = 2 * self.options.timeMultiplier,
    radius = 100,
    numberOfAttacks = 2,
    cooldown = 5 * self.options.timeMultiplier,
    movementTime = 3 * self.options.timeMultiplier,
    accuracyMultiplier = 1
  }

  return AssaultAttackPattern:new(self.anomalies:nth(1), options)
end

--- __Constructor__
--- @param anomaly Anomaly The anomaly that will be used in this phase
--- @param options? ConvergencePhaseOptions | nil A list of options to use
--- @return ConvergencePhaseOne The constructed instance
function ConvergencePhaseOne:new(anomaly, options)
  local self = classes.setmetatable(ConvergencePhase:new(Array:new({ anomaly }), ConvergencePhaseOne, options), ConvergencePhaseOne)
  return self
end

return ConvergencePhaseOne
