local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local ConvergencePhase = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhase")
local MultiAssaultAttackPattern = require("lua/explosionsNG/worldObjects/anomalies/attacks/MultiAssaultAttackPattern")

--- @class ConvergencePhaseThree : ConvergencePhase
local ConvergencePhaseThree = classes.newClass(ConvergencePhase)

--- __Static__
---
--- Creates and returns a list of options to use for the attack patterns
--- @param options ConvergencePhaseOptions The list of phase options to use
--- @return AssaultAttackPatternOptions A list of options to use for the attack patterns
--- @private
function ConvergencePhaseThree:getAttackOptions(options)
  return {
    type = "force",
    time = options.timeMultiplier,
    radius = 100,
    numberOfAttacks = 4,
    cooldown = options.timeMultiplier,
    movementTime = 2.5 * options.timeMultiplier,
    accuracyMultiplier = 0.5
  }
end

--- @return number The number of subphases
--- @see ConvergencePhase#getNumberOfSubphases
function ConvergencePhaseThree:getNumberOfSubphases()
  return 3
end

--- @return AttackPattern A new attack pattern for a new phase
--- @see ConvergencePhase#createAttackPattern
function ConvergencePhaseThree:createAttackPattern()
  local attacks = Array:new({
    { anomaly = self.anomalies:nth(1), attackOptions = ConvergencePhaseThree:getAttackOptions(self.options) },
    { anomaly = self.anomalies:nth(2), attackOptions = ConvergencePhaseThree:getAttackOptions(self.options) }
  })
  return MultiAssaultAttackPattern:new(attacks)
end

--- __Constructor__
--- @param anomalies Array<Anomaly> The anomalies that will be used in this phase
--- @param options? ConvergencePhaseOptions | nil A list of options to use
--- @return ConvergencePhaseThree The constructed instance
function ConvergencePhaseThree:new(anomalies, options)
  local self = classes.setmetatable(ConvergencePhase:new(anomalies, ConvergencePhaseThree, options), ConvergencePhaseThree)
  return self
end

return ConvergencePhaseThree
