local classes = require("lua/explosionsNG/helpers/classes")
local Array = require("lua/explosionsNG/misc/Array")
local ConvergencePhase = require("lua/explosionsNG/worldObjects/anomalies/convergence/ConvergencePhase")
local AssaultAttackPattern = require("lua/explosionsNG/worldObjects/anomalies/attacks/AssaultAttackPattern")

--- @class ConvergencePhaseTwo : ConvergencePhase
local ConvergencePhaseTwo = classes.newClass(ConvergencePhase)

--- @return number The number of subphases
--- @see ConvergencePhase#getNumberOfSubphases
function ConvergencePhaseTwo:getNumberOfSubphases()
  return 3
end

--- @return AttackPattern A new attack pattern for a new phase
--- @see ConvergencePhase#createAttackPattern
function ConvergencePhaseTwo:createAttackPattern()
  --[[--- @type AssaultAttackPatternOptions ]] local options = {
    type = math.random(2) == 1 and "dissolve" or "fire",
    time = 4 * self.options.timeMultiplier,
    radius = 100,
    numberOfAttacks = 4,
    cooldown = 0.5 * self.options.timeMultiplier,
    movementTime = 1 * self.options.timeMultiplier,
    accuracyMultiplier = 1
  }

  return AssaultAttackPattern:new(self.anomalies:nth(1), options)
end

--- __Constructor__
--- @param anomaly Anomaly The anomaly that will be used in this phase
--- @param options? ConvergencePhaseOptions | nil A list of options to use
--- @return ConvergencePhaseTwo The constructed instance
function ConvergencePhaseTwo:new(anomaly, options)
  local self = classes.setmetatable(ConvergencePhase:new(Array:new({ anomaly }), ConvergencePhaseTwo, options), ConvergencePhaseTwo)
  return self
end

return ConvergencePhaseTwo
