local AnimationStateMachine = require("lua/explosionsNG/animations/AnimationStateMachine")

-- The time (in seconds) the laser fades in
local FADE_IN_TIME = 2
-- The time (in seconds) the laser fades out in the sound file
local FADE_OUT_SOUND_TIME = 4.4

--- Fades in the laser
--- @type AnimationTimeFn<Laser>
local opening = function(laser, _, timeContext)
  laser:updateScale(timeContext.elapsedRatio)
end

--- Updates the pulsing light of the laser
--- @type AnimationFn<Laser>
local idle = function(laser, dt)
  laser.light:setField("brightness", 0, laser.pulseTimer:timeLeftRatio())
  laser.pulseTimer:update(dt)
end

--- Fades out the laser
--- @type AnimationTimeFn<Laser>
local closing = function(laser, _, timeContext)
  laser:updateScale(timeContext.timeLeftRatio)
end

--- Returns a new set of animations to animate the given laser for every state it can be in
--- @param laser Laser The laser to animate
--- @return AnimationStateMachine<LaserState, Laser> The laser animations for the given laser
return function(laser)
  local animations = --[[--- @type AnimationStateMachine<LaserState, Laser> ]] AnimationStateMachine:new(laser)
  animations:add("opening", opening, function() laser:idle() end, FADE_IN_TIME)
  animations:add("idle", idle)
  animations:add("closing", closing, function() laser:destroy() end, FADE_OUT_SOUND_TIME)
  animations:addNoop("dead")

  -- Set initial state
  animations:switch("opening")

  return animations
end