# Client/client

## Source Folders
The `ExplosionController` directory contains the main AngularJS source files. These are split into their respective
component types:
- `/components` - contains a list of directives that are controlled/stateless/dumb/presentational
- `/containers` - contains a list of directives that are uncontrolled/stateful/smart
- `/factories` - contains a list of factories that store information for this mod that are component and app agnostic
and will not be wiped unless a full client refresh occurs
- `/filters` - contains a list of filters that transform data into a better presentational format for the user

There is additionally a `types.js` file inside this directory which contains a list of type definitions for use in the
JavaScript source files.

### Compiler
It is important to note that, other than the `*.html` files, all other files inside this directory are not used by the
mod. Instead, they are all compiled into the `app.js` and `app.css` files. Within BeamNG there is no way to load up a 
new file that contains an AngularJS component, in a way that will allow the main app file to pick it up. As such, during
development, the `compiler.js` will run to collate all `*.js` and `*.css` files into the main app files. This allows all
the necessary code to be loaded by BeamNG when the app loads.

## Observables and Asynchronicity
Asynchronicity is utilised in this mod quite a bit due to needing to request data from (and send data to) the server. A
factory has been created, `engObservables`, that exposes an `Observable` class. An instance of this class acts as a 
wrapper to a particular value with can be subscribed to. When its value is updated, all subscribers will be notified of 
the change. This ensures the core data of the client app is consistent across components, including when the app is 
destroyed (e.g. when the settings menu is opened by the player). This class does not work in the traditional 
"observable" sense in that it's implementation is synchronous in nature; for more details, see the class documentation.

### Decreasing scope
A factory will not expose an observable directly; instead, it will expose the observable as of the `Subscribable<T>` 
type. This type hides the ability to change the observable's state and only allows the consumer of the factory to 
subscribe and unsubscribe from it. This is done to prevent accidental changes of a factories state directly, rather than
through that factory's various state methods.

## Server-Client Synchronising
The server is always taken as the source of truth. When the client is created/recreated, the current state of various
entities such as explosives and previews will be requested by the client so that it is aligned.

_Note that this can be broken if the main game lua is reloaded, however, this will also break the mod in doing so, so it 
is not handled by the client or server._

## Naming Conventions
To avoid conflicts in naming with existing BeamNG apps, modules, classes etc., all mod created things will be prefixed 
with the shorthand name of this mod: `eng`. 

AngularJS has a big weakness in how CSS is not encapsulated by directives at all, meaning a parent component has access
to its child's elements and can freely style them. To avoid this issue as much as possible, the selectors will be 
created using the child combinator (`>`). All CSS classes introduced by this mod will be prefixed with the shorthand 
name of the mod, followed by the name of the directive. Note this is only done for the _root element_ of the directive
to avoid extremely long selectors and to avoid reducing readability with duplicating the mod name prefix.

## Assets
All resources (such as icons and images) for the client are stored within the `/explosionsNGResources` directory. This 
also includes the compiler.

## Tests
There is a combination of manual E2E testing and automated unit/integration testing. The manual tests, aside from 
unstructured testing, manually runs through the scenarios listed in the `test-scenarios.md` file. The automated 
unit/integration testing is done using Jest, and focuses on the state management within factories.