const fs = require("fs");
const { Glob } = require("glob");

const BASE_DIRECTORY = "../modules/apps/ExplosionController";
const BLACKLIST = ["app.js", "app.css", "types.js", ".spec.js"];

const HEADER =
`/*--------------------------------------------------------------------------------------------------------------------*/
/* This is a generated file from all source files that have been compiled                                             */
/*--------------------------------------------------------------------------------------------------------------------*/

`

/**
 * Compiles a list of all files that end with the given extension into one file called `app.<extension>`. This file will
 * be placed inside the {@link BASE_DIRECTORY}.
 * @param {string} extension The file extension to compile to a single file
 */
function compile(extension) {
  const sourceFiles = new Glob(`${BASE_DIRECTORY}/**/*.${extension}`, {});
  let outputFile = HEADER;

  for (const script of sourceFiles) {
    const isBlacklisted = BLACKLIST.some(item => script.endsWith(item));
    if (!isBlacklisted) {
      const file = fs.readFileSync(script);
      outputFile += `${file}\n\n`;
    }
  }

  fs.writeFileSync(`${BASE_DIRECTORY}/app.${extension}`, outputFile);
}

compile("js");
compile("css");