/*--------------------------------------------------------------------------------------------------------------------*/
/* This is a generated file from all source files that have been compiled                                             */
/*--------------------------------------------------------------------------------------------------------------------*/

angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngRomanNumerals
   * @description Converts an integer into its corresponding roman numeral. Supports numbers from (1-5).
   */
  .filter("EngRomanNumerals", () => {
    /**
     * Converts the given integer into its corresponding roman numeral.
     * Supports value between (1, 5).
     * @param {number} input The integer to convert
     * @returns {string} The roman numeral of {@link input}
     * @throws Error if the {@link input} is not an integer between (1, 5)
     */
    function transform(input) {
      switch (input) {
        case 1: return "I";
        case 2: return "II";
        case 3: return "III";
        case 4: return "IV";
        case 5: return "V";
        default: throw new Error(`${input} is not supported by "EngRomanNumerals"`);
      }
    }

    return transform;
  });

angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngExplosiveValueText
   * @description Converts a {@link ConfiguredExplosive}'s [rank]{@link ExplosivePreset#rank} and
   * [tier]{@link ConfiguredExplosive#tier} into a title string.
   */
  .filter("EngExplosiveValueText", () => {
    /**
     * Returns a string text of the given explosive's [rank]{@link ExplosivePreset#rank} and
     * [tier]{@link ConfiguredExplosive#tier}.
     * @param {ConfiguredExplosive} explosive The explosive to get a text description for its rank and tier
     * @return {string} The text description
     */
    function transform(explosive) {
      const tier = explosive.tier;
      const rank = explosive.preset.rank;
      return `Explosive Rank ${rank} [Tier ${tier}]`;
    }

    return transform;
  });

angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngExplosiveTierDescription
   * @description Converts an integer into its corresponding explosive tier description. Supports values between (1, 4).
   */
  .filter("EngExplosiveTierDescription", () => {
    /**
     * Converts the given integer into its corresponding tier description.
     * Supports value between (1, 4).
     * @param {number} input The integer to convert
     * @returns {string} The rank description of {@link input}
     * @throws Error if the {@link input} is not an integer between (1, 4)
     */
    function transform(input) {
      switch (input) {
        case 1: return "Tier 1 explosives are just about a legal amount of explosive power. These explosives are small and more of an annoyance than damaging.";
        case 2: return "Tier 2 explosives are where the real vehicular carnage begins: more power, more range and more difficulty in trying to survive.";
        case 3: return "Tier 3 explosives are beyond dangerous. Range becomes measured in kilometres, power becomes measured in meganewtons. Even the fabric of space and time are not safe from these explosives.";
        case 4: return "Tier 4 explosives are not just apocalyptic. Explosives this powerful will annihilate anything in their path. No vehicle is too far away from them, no vehicle is strong enough to withstand them, no vehicle is fast enough to flee from them - survival is an infinitesimally small possibility.";
        default: throw new Error(`${input} is not supported by "EngExplosiveTierDescription"`);
      }
    }

    return transform;
  });

angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngExplosiveRankTitle
   * @description Converts an integer into its corresponding explosive rank title. Supports values between
   * (1, 10).
   */
  .filter("EngExplosiveRankTitle", () => {
    /**
     * Converts the given integer into its corresponding rank title.
     * Supports value between (1, 10).
     * @param {number} input The integer to convert
     * @returns {string} The rank description of {@link input}
     * @throws Error if the {@link input} is not an integer between (1, 10)
     */
    function getText(input) {
      switch (input) {
        case  1: return "Unsafe";
        case  2: return "Harmful";
        case  3: return "Hazardous";
        case  4: return "Destructive";
        case  5: return "Dangerous";
        case  6: return "Deadly";
        case  7: return "Highly Deadly";
        case  8: return "Extremely Deadly";
        case  9: return "Apocalyptic";
        case 10: return "Annihilistic";
        default: throw new Error(`${input} is not supported by "EngExplosiveRankTitle"`);
      }
    }

    return input => getText(input).toUpperCase();
  });

angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngDifficulty
   * @description Translates a given difficulty id to a corresponding effect that that difficulty will have. Note that
   * this is currently implicitly tied to the boss difficulties.
   */
  .filter("EngDifficulty", () => {
    /**
     * Translates a given difficulty id to a corresponding effect that that difficulty will have.
     * Note that this is currently implicitly tied to the boss difficulties.
     * @param {Difficulty} difficulty The integer to convert
     * @returns {string} The effect of the {@link difficulty}
     * @throws Error if the {@link difficulty} is not a valid {@link Difficulty}
     */
    function transform(difficulty) {
      switch (difficulty) {
        case "easy": return "decreased";
        case "normal": return "regular";
        case "hard": return "increased";
        default: throw new Error(`${difficulty} is not supported by "EngDifficulty"`);
      }
    }

    return transform;
  });

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engUiFactory
   * @description A store that manages states that affect how components render options and subcomponents, across
   * components. This includes things like the current view as well as extra panels that may be open.
   * @requires engObservablesFactory
   * @requires engExplosivesFactory
   */
  .factory("engUiFactory", ["engObservablesFactory", "engExplosivesFactory", (
      engObservablesFactory,
      engExplosivesFactory
  ) => {
    /** @type Observable<View> */
    const view = engObservablesFactory.newObservable("overview");
    /** @type Observable<boolean> */
    const isPositioningExplosive = engObservablesFactory.newObservable(false);
    /** @type Observable<boolean> */
    const isPlayerInAVehicle = engObservablesFactory.newObservable(false);
    /** @type Observable<boolean> */
    const isMinimised = engObservablesFactory.newObservable(false);
    /** @type {Observable<boolean>} */
    const isShowingInformation = engObservablesFactory.newObservable(false);

    /**
     * Sets the current view to the given one.
     * This function will reset the selected explosive on the {@link engExplosivesFactory} if the view has changed from
     * `"explosive"` to another view.
     * @param {View} newView The view to set
     */
    function setView(newView) {
      // Unset selected explosive if we are moving away from the configure explosive view
      if (view.value === "explosive" && newView !== "explosive") {
        engExplosivesFactory.resetSelectedExplosive();
      }

      view.next(newView);
    }

    /**
     * Sets whether an explosive is being positioned
     * @param {boolean} value The value to set
     */
    function setIsPositioningExplosive(value) {
      isPositioningExplosive.next(value);
    }

    /**
     * Sets the flag of whether the player is in a vehicle
     * @param {boolean} isInAVehicle Whether the player is now in a vehicle
     */
    function setPlayerInAVehicle(isInAVehicle) {
      isPlayerInAVehicle.next(isInAVehicle);
    }

    /**
     * Sets the flag of whether the app is minimised
     * @param {boolean} minimised Whether the app is minimised
     */
    function setIsMinimised(minimised) {
      isMinimised.next(minimised);
    }

    /**
     * Updates whether the information box is visible
     * @param {boolean} showingInformation Whether the information related to a boss is being displayed
     */
    function setIsShowingInformation(showingInformation) {
      isShowingInformation.next(showingInformation);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      view.next("overview");
      isPositioningExplosive.next(false);
      isPlayerInAVehicle.next(false);
      isShowingInformation.next(false);
    }

    return {
      view: /** @type Subscribable<View> */ view,
      isPositioningExplosive: /** @type Subscribable<boolean> */ isPositioningExplosive,
      isPlayerInAVehicle: /** @type Subscribable<boolean> */ isPlayerInAVehicle,
      isMinimised: /** @type Subscribable<boolean> */ isMinimised,
      isShowingInformation: /** @type {Subscribable<boolean>} */ isShowingInformation,
      setView,
      setIsPositioningExplosive,
      setPlayerInAVehicle,
      setIsMinimised,
      setIsShowingInformation,
      reset
    };
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engPreviewsFactory
   * @description A store that manages the state of in-game previews that are visible
   * @requires engObservablesFactory
   */
  .factory("engPreviewsFactory", ["engObservablesFactory", (engObservablesFactory) => {
    /** @type Observable<Preview[]> */
    const previews = engObservablesFactory.newObservable([]);

    /**
     * Adds the given preview to the list
     * @param {Preview} preview The preview to add
     */
    function addPreview(preview) {
      const newPreviews = [...previews.value, { ...preview }];
      previews.next(newPreviews);
    }

    /**
     * Removes the given preview to the list
     * @param {Preview} preview The preview to remove
     */
    function removePreview(preview) {
      const newPreviews = previews.value.filter(p => p !== preview);
      previews.next(newPreviews);
    }

    /**
     * Sets the list of previews to the given list
     * @param {Preview[]} newPreviews The new previews to set
     */
    function setPreviews(newPreviews) {
      previews.next([...newPreviews]);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      previews.next([]);
    }

    return {
      previews: /** @type Subscribable<Preview[]> */ previews,
      addPreview,
      removePreview,
      setPreviews,
      reset
    };
  }]);

angular.module("beamng.apps")
    /**
     * @ngdoc factory
     * @name beamng.apps:engObservablesFactory
     * @description Creates an Observable class that stores a current value that can be changed whereby any subscribers to
     * that value will be notified of the change in value. This factory exposes a single method to create new observables.
     */
    .factory("engObservablesFactory", () => {
      /**
       * An object that stores a current value and broadcasts changes in its value to subscribers. It should be noted
       * that unlike most implementations of observables, this class is synchronous. Note the example below:
       *
       * @example
       * const value = 0;
       * const observable = new Observable(0);
       * observable.subscribe(x => value = x);
       *
       * console.log(value); // Will print 0
       * observable.next(1);
       * console.log(value); // Will print 1 as `Observable.next` call is synchronous
       *
       * @template T
       */
      class Observable {
        /**
         * @constructor
         * @param {T} initialValue The value to set as the initial value of this observable
         */
        constructor(initialValue) {
          this.value = initialValue;
          /** @type {SubscriptionFn[]} */
          this.subscriptions = [];
        }

        /**
         * Sets the value of the source to the given one and emits to all subscribers
         * @param {T} newValue The value to set
         */
        next(newValue) {
          this.value = newValue;
          this.subscriptions.forEach(subscriber => subscriber(this.value));
        }

        /**
         * Adds the emitter function to this observable. Whenever this observable has a new value, the passed subscription
         * will be called with this new value.
         * This method will also call the subscription with the current value when this method is called.
         * @param {SubscriptionFn} subscription The subscription function to call when changes happen
         */
        subscribe(subscription) {
          this.subscriptions.push(subscription);
          subscription(this.value);
        }

        /**
         * Removes the emitter function from this observable
         * @param {SubscriptionFn} subscription The subscription to remove
         */
        unsubscribe(subscription) {
          this.subscriptions = this.subscriptions.filter(s => s !== subscription);
        }
      }

      /**
       * Creates a new {@link Observable} with the given initial value
       * @template T
       * @param {T} initialValue The value to set as the initial value of the observable
       * @return {Observable<T>} The created observable
       */
      function newObservable(initialValue) {
        return new Observable(initialValue);
      }

      return { newObservable };
    });

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engNavigationFactory
   * @description
   * @requies engObservablesFactory
   */
  .factory("engNavigationFactory", ["engObservablesFactory", (engObservablesFactory) => {

    /** @type {Observable<HTMLElement | null>} */
    const focusedElement = engObservablesFactory.newObservable(null);
    /** @type {Observable<HTMLElement[]>} */
    const focusableElements = engObservablesFactory.newObservable([]);

    // function getOffset(element) {
    //   const rectangle = element.getBoundingClientRect();
    //   return {
    //     left: rectangle.left + RELATIVE_RECTANGLE.left,
    //     top: rectangle.top + RELATIVE_RECTANGLE.top
    //   };
    // }

    function addNavigatableElement(element) {
      focusableElements.next([...focusableElements.value, element]);
    }

    function navigate(next = true) {
      const elements = focusableElements.value;
      const index = elements.indexOf(focusedElement.value);
      let nextIndex = next ? index + 1 : index - 1;

      if (nextIndex < 0 || nextIndex > elements.length - 1) nextIndex = 0;
      focusedElement.next(elements[nextIndex]);
    }

    function resetFocusableElements() {
      focusableElements.next([]);
    }

    function init() {
      // Get the containing element that BeamNG creates for the app
      const RELATIVE_ELEMENT = document.getElementById("eng-controller").parentElement.parentElement;

      // RELATIVE_ELEMENT.addEventListener("keydown", event => {
      //   navigate();
      // });
    }

    return {
      focusedElement: /** @type {Subscribable<HTMLElement | null>} */ focusedElement,
      addNavigatableElement,
      navigate,
      init,
      resetFocusableElements
    };
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engMiscFactory
   * @description Provides miscellaneous functions
   */
  .factory("engMiscFactory", () => {
    /**
     * Rounds a value to a given number of decimal places
     * @param {number} value The value to round
     * @param {number} decimalPlaces The number of decimal places to round to
     * @returns {number} The rounded value
     */
    function roundTo(value, decimalPlaces) {
      const power10 = Math.pow(10, decimalPlaces);
      return Math.round(value * power10) / power10;
    }

    /**
     * Checks and returns whether the given {@link value} is defined
     * @template T
     * @param {T | null | undefined} value The value to check
     * @returns {boolean} Whether the value is defined
     */
    function isDefined(value) {
      return value !== undefined && value !== null;
    }

    /**
     * Returns the {@link value} if it's defined, else the {@link defaultValue}
     * @template T
     * @param {T | null | undefined} value The value
     * @param {T} defaultValue The default value
     * @returns {T} The {@link value} if it's defined, else the {@link defaultValue}
     */
    function withDefaultValue(value, defaultValue) {
      return isDefined(value) ? value : defaultValue;
    }

    /**
     * Checks whether the given {@link position} is fully defined, which is defined as being anywhere in the coordinate
     * space that is __not__ the origin.
     * @param {Position} position The position to check
     * @returns {boolean} Whether the position is fully defined
     */
    function isPositionDefined(position) {
      return !(position.x === 0 && position.y === 0 && position.z === 0);
    }

    /**
     * Partitions an array into two subsets. The first subset contains all elements that pass the given predicate, with
     * the second subset for all other elements.
     * @template T
     * @param {T[]} array The array to partition
     * @param {(element: T) => boolean} predicate The predicate to divide the array into partitions
     * @returns {[T[], T[]]} The two partitions of the original array
     */
    function binaryPartition(array, predicate) {
      const subset1 = [];
      const subset2 = [];

      array.forEach(element => {
        const subset = predicate(element) ? subset1 : subset2;
        subset.push(element);
      })

      return [subset1, subset2];
    }

    return {
      roundTo,
      isDefined,
      withDefaultValue,
      isPositionDefined,
      binaryPartition
    };
  });

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engExplosivesFactory
   * @description A store that manages the state of all things for this mod. This includes configured explosives by the
   * user, a list of all possible explosive that can be selected and the current view that is being shown to the user.
   *
   * For components using this factory, the {@link init} method should be called as part of the component's init. This
   * will initiate all async requests and set the initial state of the factory, referred to as "synchronised". If two or
   * more components require this factory, then once the factory is synchronised, the next component will not affect
   * this factory and instead immediately return (see {@link init} for more details).
   *
   * This factory uses observables to store and manage information in this factory. To retrieve data, a component will
   * need to subscribe to the individual data by setting up a subscription (see {@link Subscribable#subscribe}).
   * Again, the {@link init} method is set up for this purpose. It is important that the component that sets up these
   * subscriptions also unsubscribes from the same data when it is destroyed, or not needed anyway (see
   * {@link Subscribable#unsubscribe}).
   * @requires engObservablesFactory
   * @requires engMiscFactory
   * @requires engEndpointsFactory
   */
  .factory("engExplosivesFactory", ["engObservablesFactory", "engMiscFactory", "engEndpointsFactory", (
      engObservablesFactory,
      engMiscFactory,
      engEndpointsFactory,
  ) => {
    let synchronised = false;
    /** @type {(() => void)[]} */
    let onSyncSubscribers = [];

    let nextId = 0;

    /** @type ExplosiveDto[] */
    let explosiveDTOs = [];
    /** @type Observable<ConfiguredExplosive[]> */
    const configuredExplosives = engObservablesFactory.newObservable([]);
    /** @type {Observable<ConfiguredExplosive | undefined>} */
    const selectedExplosive = engObservablesFactory.newObservable(undefined);
    /** @type {Observable<boolean>} */
    const lockTier4s = engObservablesFactory.newObservable(false);

    // Set up a subscription to update whether tier 4 locking should come into effect
    configuredExplosives.subscribe(checkTier4Locking);

    /**
     * Converts an {@link ExplosiveDto} object into an {@link ConfiguredExplosive} object. The returned object will have
     * its [preset]{@link ExplosivePreset} set to the corresponding preset of the given index from the {@link presets}
     * list given. The position will be set to the origin.
     * @param {ExplosiveDto} explosive The explosive DTO
     * @param {number} selectedPresetIndex The index of the presets within {@link presets} to use as the return
     * preset
     * @returns {ConfiguredExplosive} The converted object
     */
    function dtoToExplosive(explosive, selectedPresetIndex) {
      const { id, name, tier, description, properties, presets } = explosive;
      return {
        uiId: (nextId++).toString(),
        id,
        name,
        tier,
        description,
        properties,
        preset: presets[selectedPresetIndex],
        position: { x: 0, y: 0, z: 0 },
        showing: false,
        inProgress: false,
        fuse: 0
      };
    }

    /**
     * Maps an {@link ExplosiveIdentifier} to a {@link ConfiguredExplosive} object
     * @param {ExplosiveIdentifier} explosiveIdentifier The identifier to map
     * @returns {ConfiguredExplosive} The mapped configured explosive
     */
    function identifierToConfiguredExplosive(explosiveIdentifier) {
      const { uiId, explosiveId, presetId, fuse } = explosiveIdentifier;

      const { id, name, tier, description, properties, presets } = explosiveDTOs.find(({ id }) => id === explosiveId);
      /** @type ExplosivePreset */ // Not sure why type-cast is needed by Intellij?
      const preset = presets.find(({ id }) => id === presetId);
      const position = { x: 0, y: 0, z: 0 }; // This is unknown; it will never be used in the UI so shouldn't matter

      return { uiId, id, name, tier, description, properties, preset, position, fuse, showing: false, inProgress: true };
    }

    /**
     * Checks whether more than one tier 4 explosive has been configured, and sets the result to the {@link lockTier4s}
     * observable.
     * @param {ConfiguredExplosive[]} explosives The new list of configured explosive
     */
    function checkTier4Locking(explosives) {
      const shouldLock = explosives.filter(({ tier }) => tier === 4).length > 1;
      lockTier4s.next(shouldLock);
    }

    /**
     * @returns {ConfiguredExplosive} The first {@link ExplosiveDto} object in the factory, converted to a
     * {@link ConfiguredExplosive} object with the first available preset.
     */
    function createDefaultExplosive() {
      const dto = explosiveDTOs[0];
      return dtoToExplosive(dto, 0);
    }

    /**
     * Requests the list of explosives from the {@link engEndpointsFactory}.
     * This function will additionally set this factory to be synchronised.
     * @see synchronise
     */
    function initExplosives() {
      engEndpointsFactory.getListOfExplosives(explosives => {
        explosiveDTOs = explosives;

        // /** @type ConfiguredExplosive[] */
        // const createdExplosives = explosives.flatMap((explosive, explosiveIndex) => explosive.presets.map((_, index) => {
        //   if (explosiveIndex > 5) return;
        //   const position = { x: Math.random() * 100, y: Math.random() * 100, z: 0 };
        //   return { ...dtoToExplosive(explosive, index), position };
        // })).filter(engMiscFactory.isDefined);
        // const createdExplosives = Array(50).fill(null).map(() => {
        //   const position = { x: Math.random() * 100, y: Math.random() * 100, z: 0 };
        //   const explosive = explosives[Math.floor(Math.random() * (explosives.length - 1))];
        //   const presetIndex = Math.floor(Math.random() * (explosive.presets.length - 1));
        //   return { ...dtoToExplosive(explosive, presetIndex), position };
        // })
        // configuredExplosives.next(createdExplosives);

        synchronise();
      });
    }

    /**
     * Creates a new explosive with default settings and adds it to the list of configured explosives
     * @see createDefaultExplosive
     */
    function addNewConfiguredExplosive() {
      configuredExplosives.next([...configuredExplosives.value, createDefaultExplosive()]);
    }

    /**
     * Creates a {@link ConfiguredExplosive} from the given identifier and adds it to the list of configured explosives
     * @param {ExplosiveIdentifier} explosiveIdentifier The identifier to create and add an explosive for
     */
    function addConfiguredExplosiveByIdentifier(explosiveIdentifier) {
      const newConfiguredExplosive = identifierToConfiguredExplosive(explosiveIdentifier);
      configuredExplosives.next([...configuredExplosives.value, newConfiguredExplosive]);
    }

    /**
     * Removes an explosive from the list of configured explosives given its index or uiId
     * @param {number | string} property Either the index of the explosive, or the explosive uiId to remove
     */
    function removeExplosive(property) {
      if (typeof property === "number") {
        removeByIndex(property);
        return;
      }

      removeById(property);
    }

    /**
     * Removes an explosive from the list of configured explosives given its index
     * @param {number} index The index of the explosive to remove
     */
    function removeByIndex(index) {
      const explosives = configuredExplosives.value;
      const length = explosives.length;

      if (index < 0 || index >= length) {
        console.error(`Index ${index} must be positive and less than the array length (${length})`);
        return;
      }

      const explosiveToRemove = explosives[index];
      const selected = selectedExplosive.value;

      if (engMiscFactory.isDefined(selected) && selected === explosiveToRemove) {
        console.warn("Attempted to remove the currently selected explosive. Removal has been prevented");
        return;
      }

      configuredExplosives.next(explosives.filter((_, i) => i !== index));
    }

    /**
     * Removes an explosive from the list of configured explosives given its uiId
     * @param {string} id The uiId of the explosive to remove
     */
    function removeById(id) {
      const explosives = configuredExplosives.value;
      const explosiveToRemove = explosives.find(({ uiId }) => id === uiId);

      if (!engMiscFactory.isDefined(explosiveToRemove)) {
        console.warn(`Explosive with id ${id} is not present in the configuration`);
        return;
      }

      const filteredExplosives = explosives.filter(explosive => explosive !== explosiveToRemove);
      configuredExplosives.next(filteredExplosives);
    }

    /**
     * Overwrites the explosive at the given index to the given explosive. If the {@link index} is not given, then the
     * selected explosive index will be used instead.
     * @param {ConfiguredExplosive} explosive
     * @param {number} [index] The index of the explosive to set in the configuration array. Optional and if not given,
     * the selected explosive's index is used instead
     */
    function setConfiguredExplosiveByIndex(explosive, index) {
      const selectedExplosiveIndex = getSelectedExplosiveIndex();
      const explosiveIndex = engMiscFactory.withDefaultValue(index, selectedExplosiveIndex);
      const newExplosives = [...configuredExplosives.value];

      newExplosives[explosiveIndex] = explosive;
      configuredExplosives.next(newExplosives);

      // Update selected explosive if it has changed
      if (explosiveIndex === selectedExplosiveIndex) {
        selectedExplosive.next(explosive);
      }
    }

    /**
     * Sets the currently configured explosives to the given list. This will update the currently selected explosive
     * with either the updated explosive, or if it does not exist in the list, to `undefined`.
     * @param {ConfiguredExplosive[]} explosives The list of explosives to set
     */
    function setConfiguredExplosives(explosives) {
      configuredExplosives.next(explosives);

      const updatedSelectedExplosive = engMiscFactory.isDefined(selectedExplosive.value)
          ? explosives.find(({ uiId }) => uiId === selectedExplosive.value.uiId)
          : null;

      // Update selected explosive if it has changed
      if (engMiscFactory.isDefined(updatedSelectedExplosive)) {
        selectedExplosive.next(updatedSelectedExplosive);
        return
      }

      // Unset the selected explosive if it no longer exists
      selectedExplosive.next(undefined);
    }

    /**
     * Overwrites the preset of the currently selected explosive given the index of the new preset to set
     * @param {number} presetIndex The index of the preset to set
     */
    function setSelectedPresetByIndex(presetIndex) {
      const presets = getSelectedExplosiveDto().presets;
      const preset = presets[presetIndex];
      const updatedExplosive = /** @type ConfiguredExplosive */ { ...selectedExplosive.value, preset };

      setConfiguredExplosiveByIndex(updatedExplosive);
      selectedExplosive.next(updatedExplosive);
    }
    /**
     * Sets the currently selected explosive to the explosive that exists in the configured explosives with the given
     * index
     * @param {number} index The index of the list of currently configured explosives to get the new selected explosive
     */
    function setSelectedExplosiveByIndex(index) {
      const newSelectedExplosive = configuredExplosives.value[index];
      selectedExplosive.next(newSelectedExplosive);
    }

    /**
     * Resets the currently selected explosive to `undefined`
     */
    function resetSelectedExplosive() {
      selectedExplosive.next(undefined);
    }

    /**
     * @returns {number} The index of the currently selected explosive in the list of configured explosives
     */
    function getSelectedExplosiveIndex() {
      return configuredExplosives.value.findIndex(explosive => explosive === selectedExplosive.value);
    }

    /**
     * @returns {ExplosiveDto} The DTO of the currently selected explosive;
     */
    function getSelectedExplosiveDto() {
      const configurationId = selectedExplosive.value.id;
      return explosiveDTOs.find(({ id }) => id === configurationId);
    }

    /**
     * @returns {ExplosiveDto[]} The list of explosives available, in DTO format
     */
    function getExplosiveDTOs() {
      return [...explosiveDTOs];
    }

    /**
     * Sets this factory to {@link synchronised}. This function will also call each of the {@link onSyncSubscribers},
     * before emptying the array.
     */
    function synchronise() {
      synchronised = true;
      onSyncSubscribers.forEach(subscriber => subscriber());
      onSyncSubscribers = [];
    }

    /**
     * Merges the current list of configured explosives with the given list of explosive identifiers from the server.
     * The list of configured explosives is updated in the following ways:
     *  - If the explosive is not marked as detonating/in progress in this factory, then it remains in the list
     *  - If the explosive is marked as detonating/in progress in this factory, and it exists in the server list
     *  too, then it also remains in the list
     *  - If the explosive is marked as detonating/in progress in this factory, but does not exist in the server list,
     *  then it is removed
     *  - If the explosive exists **only** on the server, then it is mapped to a {@link ConfiguredExplosive} and added
     *  to the list
     * @param {ExplosiveIdentifier[]} serverExplosives The list of explosive identifiers from the server
     */
    function softSyncWithServer(serverExplosives) {
      const serverUiIds = serverExplosives.map(({ uiId }) => uiId);

      // Keep any explosive on the factory that is either not in progress, or also exists on the server
      const existingExplosives = configuredExplosives.value.filter(({ uiId, inProgress }) => !inProgress || serverUiIds.includes(uiId));

      const serverOnlyExplosives = serverExplosives
          .filter(({ uiId }) => !configuredExplosives.value.find(e => e.uiId === uiId))
          .map(identifierToConfiguredExplosive);

      // Keep the order from the factory, and add new explosives to the end of the list
      const currentExplosives = [...existingExplosives, ...serverOnlyExplosives];
      configuredExplosives.next(currentExplosives);
    }

    /**
     * Creates and sets the list of configured explosives, mapped from a list of explosive identifiers from the server.
     * This function will overwrite all currently configured explsovies, included the currently selected explosive.
     * Note that this function is not necessarily synchronous; the given callback function will be called once this
     * function has completed.
     * @param {ExplosiveIdentifier[]} serverExplosives The list of explosive identifiers from the server
     * @param {() => void} callback The callback function to call once done
     */
    function hardSyncWithServer(serverExplosives, callback) {
      // Ensure that the DTOs are available for use
      init(() => {
        const serverConfiguredExplosives = serverExplosives.map(identifierToConfiguredExplosive);

        selectedExplosive.next(undefined);
        configuredExplosives.next(serverConfiguredExplosives);

        callback();
      });
    }

    /**
     * Initialises this factory by performing all async requests it needs. Once these have completed, the given callback
     * function will be called. This should typically be used to set up subscriptions.
     * If this factory has already performed its async requests, then this method will immediately call the callback.
     * @param {() => void} onSync The callback to call after synchronisation has completed
     */
    function init(onSync) {
      if (synchronised) {
        onSync();
        return;
      }

      initExplosives();
      onSyncSubscribers.push(onSync);
    }

    return {
      configuredExplosives: /** @type Subscribable<ConfiguredExplosive[]> */ configuredExplosives,
      selectedExplosive: /** @type Subscribable<ConfiguredExplosive> */ selectedExplosive,
      lockTier4s: /** @type Subscribable<boolean> */ lockTier4s,
      dtoToExplosive,
      addNewConfiguredExplosive,
      addConfiguredExplosiveByIdentifier,
      removeExplosive,
      setConfiguredExplosiveByIndex,
      setConfiguredExplosives,
      setSelectedPresetByIndex,
      setSelectedExplosiveByIndex,
      resetSelectedExplosive,
      getExplosiveDTOs,
      softSyncWithServer,
      hardSyncWithServer,
      init
    }
  }
]);

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engEndpointsFactory
   * @description Provides an API to make requests to the ExplosiveNG server extension, both retrieving data from and
   * sending data to this server.
   */
  .factory("engEndpointsFactory", () => {
    // Location of the server
    const EXTENSION = "extensions.explosionsNG";

    /**
     * The `bngApi` incorrectly returns an empty object instead of an empty array when returned from the server.
     * This function corrects the type.
     * @template T
     * @param {{} | T[]} arrayOrObject The array or object returned from the `bngApi`
     * @returns {T[]} The fixed array
     */
    function fixBngApiArray(arrayOrObject) {
      return Array.isArray(arrayOrObject) ? arrayOrObject : [];
    }

    /**
     * Makes a request to the server extension to get the current state on the server side
     * @param {(response: StateResponse) => void} callback The function to run with the server response. The server will
     * return a {@link StateResponse}.
     */
    function loadState(callback) {
      bngApi.engineLua(`${EXTENSION}.getCurrentState()`, state => {
        callback(/** @type StateResponse */ { ...state, explosives: fixBngApiArray(state.explosives) });
      });
    }

    /**
     * Makes a request to the server extension with the given combined list of explosives to create and place
     * @param {PlaceExplosiveRequest[]} explosives The explosives to create and place on the server response
     */
    function placeExplosives(explosives) {
      bngApi.engineLua(`${EXTENSION}.placeExplosives(${bngApi.serializeToLua(explosives)})`);
    }

    /**
     * Makes a request to the server extension for the list of explosives available. The given {@link callback} will be
     * called on server response.
     * @param {(explosives: ExplosiveDto[]) => void} callback The function to run with the server response. The server
     * will return the list of all explosives that can be created by this mod.
     */
    function getListOfExplosives(callback) {
      bngApi.engineLua(`${EXTENSION}.getListOfExplosives()`, response => callback(fixBngApiArray(response)));
    }

    /**
     * Makes a request to the server extension to create a new preview at the given {@link position}. The given
     * {@link callback} will be called on server response.
     * @param {string} explosiveId The explosive id
     * @param {string} presetId The preset's id of the explosive
     * @param {Position} position The position to use
     * @param {(id: string) => void} callback The function to run with the server response. The server will return the
     * id of the created preview
     */
    function createPreviewByPosition(position, explosiveId, presetId, callback) {
      bngApi.engineLua(
          `${EXTENSION}.createPreviewByPosition(${bngApi.serializeToLua(position)}, "${explosiveId}", "${presetId}")`,
          callback
      );
    }

    /**
     * Makes a request to the server extension to create a new preview with the {@link option} to set its position. The
     * given {@link callback} will be called on server response.
     * @param {AutoCoordinateInputType} option The coordinate type to use as the position
     * @param {string} explosiveId The explosive id
     * @param {string} presetId The preset's id of the explosive
     * @param {(response: { id: string, position: Position }) => void} callback The function to run with the server
     * response. The server will return the id of the created preview and its position it was placed at.
     */
    function createPreviewByOption(option, explosiveId, presetId, callback) {
      bngApi.engineLua(`${EXTENSION}.createPreviewByOption("${option}", "${explosiveId}", "${presetId}")`, callback);
    }

    /**
     * Makes a request to the server extension with the given {@link position} to update the preview position
     * @param {string} id The id of the preview to update
     * @param {Position} position The position to use
     */
    function updatePreviewPosition(id, position) {
      bngApi.engineLua(`${EXTENSION}.updatePreviewPosition("${id}", ${bngApi.serializeToLua(position)})`);
    }

    /**
     * Makes a request to the server extension with the given {@link option} to update the preview position. The given
     * {@link callback} will be called on server response.
     * @param {string} id The id of the preview to update
     * @param {AutoCoordinateInputType} option The coordinate type to use as the position
     * @param {(response: { id: string, position: Position }) => void} callback The function to run with the server
     * response. The server will return the position at which the preview was placed at
     */
    function updatePreviewPositionByType(id, option, callback) {
      bngApi.engineLua(`${EXTENSION}.updatePreviewPositionByType("${id}", "${option}")`, callback);
    }

    /**
     * Makes a request to the server extension to delete the displayed preview for the preview that matches the given
     * {@link id}.
     * @param {string} id The id of the preview to delete
     */
    function deletePreview(id) {
      bngApi.engineLua(`${EXTENSION}.deletePreview("${id}")`);
    }

    /**
     * Makes a request to the server extension to destroy all active explosives
     */
    function destroyExplosives() {
      bngApi.engineLua(`${EXTENSION}.destroyAll()`);
    }

    /**
     * Makes a request to the server extension to destroy all active explosives
     */
    function destroyExplosive(uiId) {
      bngApi.engineLua(`${EXTENSION}.destroyExplosive("${uiId}")`);
    }

    /**
     * Makes a request to the server extension to start chaos mode with the given configuration
     * @param {ChaosModeConfiguration} configuration The configuration
     */
    function startChaosMode(configuration) {
      bngApi.engineLua(`${EXTENSION}.startChaosMode(${bngApi.serializeToLua(configuration)})`)
    }

    /**
     * Makes a request to the server extension to stop chaos mode
     */
    function stopChaosMode() {
      bngApi.engineLua(`${EXTENSION}.stopChaosMode()`);
    }

    return {
      loadState,
      placeExplosives,
      getListOfExplosives,
      createPreviewByPosition,
      createPreviewByOption,
      updatePreviewPosition,
      updatePreviewPositionByType,
      deletePreview,
      destroyExplosives,
      destroyExplosive,
      startChaosMode,
      stopChaosMode
    };
  });

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engChaosModeFactory
   * @description A store that manages the state of the chaos mode
   * @requies engObservablesFactory
   */
  .factory("engChaosModeFactory", ["engObservablesFactory", "engExplosivesFactory", (engObservablesFactory, engExplosivesFactory) => {
    /** @type {Observable<ExplosiveDto[]>} */
    const explosiveOptions = engObservablesFactory.newObservable([]);
    /** @type {Observable<ChaosModeConfiguration>} */
    const configuration = engObservablesFactory.newObservable(getDefaultConfiguration());
    /** @type {Observable<boolean>} */
    const isActive = engObservablesFactory.newObservable(false);
    /**
     * Note that this is the normalised timer value
     * @type {Observable<number | null>}
     */
    const timer = engObservablesFactory.newObservable(null);

    /**
     * @return {ChaosModeConfiguration} The default configuration
     */
    function getDefaultConfiguration() {
      return {
        maximumConcurrentExplosives: 5,
        interval: 15,
        placementRandomness: 50,
        allowedExplosives: explosiveOptions.value.map(({ id }) => id)
      };
    }

    /**
     * Sets whether the mode is active or not
     * @param {boolean} newValue The new active whether
     */
    function setActive(newValue) {
      isActive.next(newValue);
    }

    /**
     * Sets the timer
     * @param {number | null} newValue The new value to set
     */
    function setTimer(newValue) {
      timer.next(newValue);
    }

    /**
     * Sets the current configuration
     * @param {ChaosModeConfiguration} newConfiguration The configuration to set
     */
    function setConfiguration(newConfiguration) {
      configuration.next(newConfiguration);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      configuration.next(getDefaultConfiguration());
      isActive.next(false);
    }

    /**
     * Initialises the {@link engExplosivesFactory} and populates the {@link explosiveOptions} with valid explosives (
     * all except tier 4 explosives).
     * This function then updates the {@link configuration} with the default configuration.
     */
    function init() {
      engExplosivesFactory.init(() => {
        // Disallow tier 4 explosives
        const allowedOptions = engExplosivesFactory.getExplosiveDTOs().filter(({ tier }) => tier < 4);
        explosiveOptions.next(allowedOptions);
        // Update from above changes
        configuration.next(getDefaultConfiguration());
      });
    }

    init();

    return {
      configuration: /** @type {Subscribable<ChaosModeConfiguration>} */ configuration,
      explosiveOptions: /** @type {Subscribable<ExplosiveDto[]>} */ explosiveOptions,
      isActive: /** @type {Subscribable<boolean>} */ isActive,
      timer: /** @type {Subscribable<number | null>} */ timer,
      setActive,
      setTimer,
      setConfiguration,
      reset
    };
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engBossesFactory
   * @description A store that manages the state of active bosses. This is currently specifically tailored towards the
   * "The Convergence" explosive.
   * @requies engObservablesFactory
   */
  .factory("engBossesFactory", ["engObservablesFactory", engObservablesFactory => {
    /** @type {Observable<ConvergencePhaseProgress[] | null>} */
    const bossBattleProgress = engObservablesFactory.newObservable(null);
    /** @type {Observable<boolean>} */
    const hasFailedSegment = engObservablesFactory.newObservable(false);

    /**
     * Updates the current boss battle progress with the given progress list
     * @param {ConvergencePhaseProgress[] | null} newProgress The new progress of the boss
     */
    function setBossBattleProgress(newProgress) {
      bossBattleProgress.next(newProgress);
    }

    /**
     * Updates whether the player has failed the current segment of the boss
     * @param {boolean} failed Whether the player has failed the current segment of the boss
     */
    function setHasFailedSegment(failed) {
      hasFailedSegment.next(failed);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      bossBattleProgress.next(null);
      hasFailedSegment.next(false);
    }

    return {
      bossBattleProgress: /** @type {Subscribable<ConfiguredExplosive[] | null>} */ bossBattleProgress,
      hasFailedSegment: /** @type {Subscribable<boolean>} */ hasFailedSegment,
      setBossBattleProgress,
      setHasFailedSegment,
      reset
    };
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engNavigatable
   * @description
   * @requires engNavigationFactory
   */
  .directive("engNavigatable", ["engNavigationFactory", engNavigationFactory => {
    return {
      link: ($scope, element) => {
        const THIS_ELEMENT = angular.element(element.children())[0];

        const focusedElementSubscription = focusedElement => {
          const isFocused = focusedElement === THIS_ELEMENT;
          THIS_ELEMENT.style.border = isFocused ? "2px solid white" : "";
        };

        function init() {
          engNavigationFactory.addNavigatableElement(THIS_ELEMENT);
          engNavigationFactory.focusedElement.subscribe(focusedElementSubscription);
        }

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engNavigationFactory.focusedElement.unsubscribe(focusedElementSubscription);
        });

        init();
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engOverviewContainer
   * @description A container component that displays an overview for every configured explosive by the user
   * @requires engExplosivesFactory
   * @requires engNavigationFactory
   * @requires engEndpointsFactory
   * @requires engChaosModeFactory
   * @requires engPreviewsFactory
   * @requires engMiscFactory
   * @requires engUiFactory
   */
  .directive("engOverviewContainer", ["engExplosivesFactory", "engNavigationFactory", "engEndpointsFactory", "engChaosModeFactory", "engPreviewsFactory", "engMiscFactory", "engUiFactory", "$timeout", (
      engExplosivesFactory,
      engNavigationFactory,
      engEndpointsFactory,
      engChaosModeFactory,
      engPreviewsFactory,
      engMiscFactory,
      engUiFactory,
      $timeout
  ) => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/containers/overview-container/overview-container.html",
      restrict: "E",
      scope: false,
      controller: $scope => {
        /** @type ConfiguredExplosive[] */
        $scope.configuredExplosives = [];
        /** @type Preview[] */
        $scope.previews = [];
        $scope.isDetonateDisabled = true;
        $scope.isMinimised = false;

        $scope.chaosModeTimer = null;

        /**
         * This is to get around the classic funkiness of AngularJS and the state inside here not necessary updating
         * the template HTML.
         *
         * This will be set from within a {@link $timeout} call so that the explosives are reflected in the DOM.
         *
         * This approach has been taken, rather than just doing this on the {@link configuredExplosives} themselves,
         * as a lot of functions inside here rely on the synchronous nature of the `Observable` class and this would be
         * broken if it were to be wrapped in a {@link $timeout} call.
         *
         * This list will be maintained and will store the exactly same reference to the {@link configuredExplosives}.
         * @type ConfiguredExplosive[]
         **/
        $scope.templateExplosives = $scope.configuredExplosives;

        /**
         * Sets the current list of previews to the given previews
         * @param {Preview[]} previews The next list of previews
         */
        const previewsSubscription = previews => { $scope.previews = previews; };

        /**
         * Sets whether the app is currently minimise to the new given value
         * @param {boolean} isMinimised The new state
         */
        const isMinimisedSubscription = isMinimised => { $scope.isMinimised = isMinimised; };

        /**
         * Sets the chaos mode timer to the new given value
         * @param {number | null} timer The new timer value
         */
        const chaosModeTimerSubscription = timer => { $scope.chaosModeTimer = timer; };

        /**
         * Sets the current configured explosives to the given list of {@link explosives}.
         * Updates whether the detonation button is disabled.
         * @param {ConfiguredExplosive[]} explosives The new value of configured explosives
         * @see isDetonationAllowed
         */
        const configuredExplosivesSubscription = explosives => {
          $scope.configuredExplosives = explosives;
          $scope.isDetonateDisabled = !isDetonationAllowed(explosives);

          $timeout(() => {
            $scope.templateExplosives = explosives;
          });
        };

        /**
         * Sets the currently selected index to the given one before going to the explosive's configuration view
         * @param {number} index The index of the explosive clicked
         */
        function goToExplosiveViewByIndex(index) {
          engExplosivesFactory.setSelectedExplosiveByIndex(index);
          engUiFactory.setView("explosive");
        }

        /**
         * Deletes an in-game preview using the {@link engEndpointsFactory}, given the preview, and additionally removes
         * it from the {@link engPreviewsFactory}.
         * @param {Preview} preview The preview to remove
         */
        function requestPreviewDeletion(preview) {
          engEndpointsFactory.deletePreview(preview.previewId);
          engPreviewsFactory.removePreview(preview);
        }

        /**
         * Deletes an in-game preview using the {@link engEndpointsFactory} for a given explosive
         * @param {ConfiguredExplosive} explosive The explosive to delete the preview for
         */
        function deletePreview(explosive) {
          const preview = $scope.previews.find(({ explosiveUiId }) => explosiveUiId === explosive.uiId);

          if (!engMiscFactory.isDefined(preview)) {
            console.error("Cannot delete preview for this explosive as it does not exist");
            return;
          }

          requestPreviewDeletion(preview);
        }

        /**
         * Converts a {@link ConfiguredExplosive} to a {@link PlaceExplosiveRequest} object by extracting out the
         * various ids and properties.
         * @param {ConfiguredExplosive} explosive The explosive
         * @returns {PlaceExplosiveRequest} The request from the explosive
         */
        function explosiveToRequest(explosive) {
          const { uiId, id, preset, position, fuse } = explosive;
          return { uiId, position, fuse, explosiveId: id, presetId: preset.id };
        }

        /**
         * Returns a new explosive from the given explosive where the [showing]{@link ConfiguredExplosive#showing} flag
         * is set to `false`, and the [inProgress]{@link ConfiguredExplosive#inProgress} flag is set to `true`.
         * @param {ConfiguredExplosive} explosive The explosive to set
         * @returns {ConfiguredExplosive} The updated explosive
         */
        function setExplosiveDetonationState(explosive) {
          return explosive.inProgress ? explosive : { ...explosive, showing: false, inProgress: true };
        }

        /**
         * Returns whether the given list of explosives is populated with at least one explosive, and that there is at
         * least one explosive that is not yet in progress.
         * @param {ConfiguredExplosive[]} explosives The list of explosives to check
         * @returns {boolean} Whether there is at least one explosive and that at least one is not yet in progress
         */
        function isDetonationAllowed(explosives) {
          return explosives.length > 0 && explosives.some(({ inProgress }) => !inProgress);
        }

        /**
         * Checks whether a preview exists for the configured explosive at the given index. If it does, this function
         * will then request its deletion and update the explosive on the {@link engExplosivesFactory} with it's
         * `showing` flag set to `false`.
         * @param {number} index The index of the explosive to attempt to delete a preview of
         */
        function removeExistingPreview(index) {
          const explosive = $scope.configuredExplosives[index];
          const preview = $scope.previews.find(({ explosiveUiId }) => explosiveUiId === explosive.uiId);

          if (engMiscFactory.isDefined(preview)) {
            requestPreviewDeletion(preview);
            engExplosivesFactory.setConfiguredExplosiveByIndex({ ...explosive, showing: false }, index);
          }
        }

        /**
         * Deletes the preview from each explosive, if they are being previewed.
         * Note that this will only update the internal list of configured explosives, and the
         * {@link engExplosivesFactory}, if the {@link updateState} flag is set to `true`.
         * @param {boolean} [updateState] Whether to update the state as well. Defaults to `false`.
         */
        function removeAllPreviewsFromExplosives(updateState = false) {
          $scope.configuredExplosives
              .filter(({ showing }) => showing)
              .forEach(deletePreview);

          if ($scope.previews.length > 0) {
            console.warn("There are still previews shown but no explosives are being shown currently. This is a de-sync warning.");
          }

          if (updateState) {
            const updatedExplosives = $scope.configuredExplosives.map(explosive => /** @type ConfiguredExplosive */ ({ ...explosive, showing: false }));
            engExplosivesFactory.setConfiguredExplosives(updatedExplosives);
          }
        }

        /**
         * Toggles whether the app is minimised on the {@link engUiFactory}
         */
        function onAppMinimise() {
          const isNowMinimised = !$scope.isMinimised;
          engUiFactory.setIsMinimised(isNowMinimised);
        }

        /**
         * Sets the view to the `"chaos"` view on the {@link engUiFactory}
         */
        function onChaosModeClicked() {
          engUiFactory.setView("chaos");
        }

        /**
         * Sets the currently selected index to the given one before going to the explosive's configuration positioning
         * view.
         *
         * If the explosive is currently in progress, this function will do nothing.
         * @param {number} index The index of the explosive clicked
         */
        function onOpenClick(index) {
          if ($scope.configuredExplosives[index].inProgress) return;

          removeAllPreviewsFromExplosives(true);
          goToExplosiveViewByIndex(index);
          engUiFactory.setIsPositioningExplosive(true);
        }

        /**
         * Adds a new explosive to the list of configured explosives, before selecting and going to this newly created
         * explosive's configuration view.
         */
        function onAddNewExplosiveClick() {
          engExplosivesFactory.addNewConfiguredExplosive();
          removeAllPreviewsFromExplosives(true);
          goToExplosiveViewByIndex($scope.configuredExplosives.length - 1);
        }

        /**
         * Removes an explosive given its index. If the explosive is current in progress, then this function will
         * request its destruction on the {@link engEndpointsFactory}. If not, and a preview is being shown for this
         * explosive, this function will remove the in-game preview.
         * @param {number} index The index of the explosive to remove/destroy
         */
        function onDeleteClick(index) {
          const { inProgress, uiId } = $scope.configuredExplosives[index];

          if (inProgress) {
            engEndpointsFactory.destroyExplosive(uiId);
            // Note that we do not need to remove it from the engExplosivesFactory - the server will alert the UI when
            // the above end point is called and the server destroys the explosive on its end.
            return;
          }

          removeExistingPreview(index);
          engExplosivesFactory.removeExplosive(index);
        }

        /**
         * Places the explosives using the {@link engEndpointsFactory} endpoint with the configured explosives.
         * This function will then hide and delete all previews on the {@link engPreviewsFactory}, before also updating
         * the configured explosives on the {@link engExplosivesFactory}.
         */
        function onDetonate() {
          // Place all non-detonated explosives
          const explosiveRequests = $scope.configuredExplosives
              .filter(({ inProgress }) => !inProgress)
              .map(explosiveToRequest);
          engEndpointsFactory.placeExplosives(explosiveRequests);

          // Delete all previews, but do not update state as more modifications need to be made to the explosives
          removeAllPreviewsFromExplosives();

          // Update the factory with the newly detonated explosives
          const updatedExplosives = $scope.configuredExplosives.map(setExplosiveDetonationState);
          engExplosivesFactory.setConfiguredExplosives(updatedExplosives);
        }

        /**
         * Toggles whether the explosive at the given {@link index} is showing or not. If the explosive is no longer
         * being shown then this function will delete the in-game preview. Otherwise, this function will create it.
         * @param {number} index The index of the explosive to show/hide preview
         */
        function onPreviewClick(index) {
          const explosive = $scope.configuredExplosives[index];
          const newExplosive = { ...explosive, showing: !explosive.showing };

          engExplosivesFactory.setConfiguredExplosiveByIndex(newExplosive, index);

          if (!newExplosive.showing) {
            deletePreview(newExplosive);
            return;
          }

          const { id, uiId, position, preset } = newExplosive;
          engEndpointsFactory.createPreviewByPosition(position, id, preset.id, previewId => {
            engPreviewsFactory.addPreview({ previewId, explosiveUiId: uiId });
          });
        }

        /**
         * Initialises the component.
         * Initialises the {@link engExplosivesFactory}, and subscribes to various observables on various factories.
         * Additionally, this function will reset the focusable elements on the {@link engNavigationFactory}.
         */
        function init() {
          engNavigationFactory.resetFocusableElements();

          engExplosivesFactory.init(() => {
            engExplosivesFactory.configuredExplosives.subscribe(configuredExplosivesSubscription);
          });

          engUiFactory.isMinimised.subscribe(isMinimisedSubscription);
          engPreviewsFactory.previews.subscribe(previewsSubscription);
          engChaosModeFactory.timer.subscribe(chaosModeTimerSubscription);
        }

        $scope.onAppMinimise = onAppMinimise;
        $scope.onChaosModeClicked = onChaosModeClicked;
        $scope.onOpenClick = onOpenClick;
        $scope.onAddNewExplosiveClick = onAddNewExplosiveClick;
        $scope.onDeleteClick = onDeleteClick;
        $scope.onDetonate = onDetonate;
        $scope.onPreviewClick = onPreviewClick;

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engExplosivesFactory.configuredExplosives.unsubscribe(configuredExplosivesSubscription);
          engPreviewsFactory.previews.unsubscribe(previewsSubscription);
          engUiFactory.isMinimised.unsubscribe(isMinimisedSubscription);
          engChaosModeFactory.timer.unsubscribe(chaosModeTimerSubscription);
        });

        init();
      }
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engTooltip
   * @description Acts as a wrapper to an element, displaying a tooltip when hovering over the transcluded element
   * passed to it. This tooltip can additionally display an optional title before the text.
   * @param {string} text The main text to show in te tooltip
   * @param {string} [title] An optional title to display before the text
   * @param {"info" | "non-applicable"} [variant] The variant of the tooltip. Defaults to `"info"`.
   * @param {boolean} [show] Whether to show the tooltip or not. Defaults to `true`.
   */
  .directive("engTooltip", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/tooltip/tooltip.html",
      restrict: "E",
      transclude: true,
      scope: {
        title: "<",
        text: "<",
        show: "<",
        variant: "<"
      },
      link: $scope => {
        $scope.variant = $scope.variant || "info";
        $scope.show = $scope.show !== undefined ? $scope.show : true;
      }
    };
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engController
   * @description The app container that is created by BeamNG when the app is added to the game. Acts as the root
   * element that is responsible for what view is shown but does not control the behaviour of these views. This acts as
   * an AngularJS controller, but is defined as a directive (as it must in order to work correctly as a BeamNG app).
   * @requires engExplosivesFactory
   * @requires engNavigationFactory
   * @requires engChaosModeFactory
   * @requires engEndpointsFactory
   * @requires engPreviewsFactory
   * @requires engMiscFactory
   * @requires engUiFactory
   */
  .directive("engController", ["engExplosivesFactory", "engNavigationFactory", "engChaosModeFactory", "engEndpointsFactory", "engPreviewsFactory", "engBossesFactory", "engMiscFactory", "engUiFactory", (
      engExplosivesFactory,
      engNavigationFactory,
      engChaosModeFactory,
      engEndpointsFactory,
      engPreviewsFactory,
      engBossesFactory,
      engMiscFactory,
      engUiFactory,
  ) => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/containers/main-container/main-container.html",
      restrict: "E",
      scope: false,
      controller: $scope => {
        /** @type {View | undefined} */
        $scope.view = undefined;
        /** @type ConfiguredExplosive[] */
        $scope.configuredExplosives = [];
        $scope.showAnimations = true;
        $scope.areMaterialsSetUp = true;

        /**
         * Sets the current view to the new given view.
         * @param {View} view The new view
         */
        const viewSubscription = view => { $scope.view = view; };

        /**
         * Sets the currently configured explosives to the new given list
         * @param {ConfiguredExplosive[]} explosives The new explosives
         */
        const configuredExplosivesSubscription = explosives => { $scope.configuredExplosives = explosives; };

        /**
         * Removes the explosive withe given uiId from the {@link engExplosivesFactory}
         * @param {string} explosiveUiId The uiId of the explosive
         */
        function onExplosiveFinished(explosiveUiId) {
          const explosiveToDelete = $scope.configuredExplosives.find(({ uiId }) => uiId === explosiveUiId);

          if (!engMiscFactory.isDefined(explosiveToDelete)) {
            console.warn(`Explosive with id ${explosiveUiId} has finished but is not present in the configuration`);
            return;
          }

          engExplosivesFactory.removeExplosive(explosiveUiId);
        }

        /**
         * Updates the {@link engBossesFactory}'s state with the given boss data
         * @param {BossData | undefined} bossData The data from the currently active boss
         */
        function updateBossFactoryState(bossData) {
          const hasActiveBoss = engMiscFactory.isDefined(bossData);
          engBossesFactory.setBossBattleProgress(hasActiveBoss ? bossData.progress : null);
          engBossesFactory.setHasFailedSegment(hasActiveBoss && bossData.failed);
        }

        /**
         * Sets whether the chaos mode is active, and the current state on the {@link engChaosModeFactory}
         * @param {ChaosModeConfiguration | undefined} configuration The new chaos mode configuration
         */
        function updateChaosModeState(configuration) {
          const isChaosModeActive = engMiscFactory.isDefined(configuration);
          engChaosModeFactory.setActive(isChaosModeActive);
          engChaosModeFactory.setTimer(isChaosModeActive ? configuration.timer : null);
          if (isChaosModeActive) engChaosModeFactory.setConfiguration(configuration);
        }

        /**
         * Merges the state's various factories with the given data from the server. This function will set whether the
         * player is in a vehicle on the {@link engUiFactory}, the boss data on the {@link engBossesFactory}, the chaos
         * mode state on the {@link engChaosModeFactory}, and combine the configured explosives on the
         * {@link engExplosivesFactory}.
         * @param {ExplosiveIdentifier[]} explosives The list of explosive identifiers on the server
         * @param {BossData | undefined} bossData The data from the currently active boss
         * @param {boolean} isPlayerInAVehicle Whether the player is currently in a vehicle
         * @param {ChaosModeConfiguration | undefined} chaosMode The current chaos mode configuration
         */
        function updateStateFromServer(explosives, bossData, isPlayerInAVehicle, chaosMode) {
          engUiFactory.setPlayerInAVehicle(isPlayerInAVehicle);
          engExplosivesFactory.softSyncWithServer(explosives);
          updateBossFactoryState(bossData);
          updateChaosModeState(chaosMode);
        }

        /**
         * Resets and sets the states of various factories with the given data from the server. This function will set
         * whether the player is in a vehicle on the {@link engUiFactory}, the boss data on the {@link engBossesFactory}
         * , the chaos mode state on the {@link engChaosModeFactory}, and set the configured explosives on the
         * {@link engExplosivesFactory}.
         *
         * The {@link callback} will be called once the states have been successfully updated.
         * @param {ExplosiveIdentifier[]} explosives The list of explosive identifiers on the server
         * @param {BossData | undefined} bossData The data from the currently active boss
         * @param {boolean} isPlayerInAVehicle Whether the player is currently in a vehicle
         * @param {ChaosModeConfiguration | undefined} chaosMode The current chaos mode configuration
         * @param {() => void} callback A callback function to call once the factories have been set
         */
        function resetStateToServer(explosives, bossData, isPlayerInAVehicle, chaosMode, callback) {
          engUiFactory.reset();
          engBossesFactory.reset();
          engPreviewsFactory.reset();
          engChaosModeFactory.reset();

          engUiFactory.setPlayerInAVehicle(isPlayerInAVehicle);
          updateBossFactoryState(bossData);
          updateChaosModeState(chaosMode);

          engExplosivesFactory.hardSyncWithServer(explosives, callback);
        }

        /**
         * Subscribe to all observables used by this component
         */
        function initSubscriptions() {
          engUiFactory.view.subscribe(viewSubscription);
          engExplosivesFactory.init(() => {
            engExplosivesFactory.configuredExplosives.subscribe(configuredExplosivesSubscription);
          });
        }

        /**
         * Initialises the component.
         * Calls the server to get the current state of the mod. If the UI needs a full reset, then this function will
         * reset the UI's state to that server's state (see {@link resetStateToServer}). Otherwise, this function will
         * combine the UI's current state with this server state (see {@link updateStateFromServer}).
         * Both functions will initialise the subscriptions.
         * @see resetStateToServer
         * @see updateStateFromServer
         */
        function init() {
          // engNavigationFactory.init();

          engEndpointsFactory.loadState(({
            explosives,
            bossData,
            isPlayerInAVehicle,
            needsReset,
            areMaterialsSetUp,
            chaosMode
          }) => {
            $scope.areMaterialsSetUp = areMaterialsSetUp;

            if (needsReset) {
              resetStateToServer(explosives, bossData, isPlayerInAVehicle, chaosMode, initSubscriptions);
              $scope.showAnimations = true;
              return;
            }

            updateStateFromServer(explosives, bossData, isPlayerInAVehicle, chaosMode);
            initSubscriptions();

            $scope.showAnimations = false;
            /**
             * We want to "flicker" the showAnimations so that it does not show animations initially, but does
             * subsequently after.
             * A little bit hackish, but works really nicely and simply.
             */
            setTimeout(() => $scope.showAnimations = true, 300);
          });
        }

        /**
         * An explosive has finished on the server. Removes the explosive from the {@link engExplosivesFactory}.
         */
        $scope.$on("engExplosiveFinished", (_event, id) => onExplosiveFinished(/** @type string */ id));

        /**
         * An explosive was added on the server. Adds the explosive to the {@link engExplosivesFactory}.
         */
        $scope.$on("engExplosiveAdded", (_event, id) => engExplosivesFactory.addConfiguredExplosiveByIdentifier(/** @type ExplosiveIdentifier */ id));

        /**
         * The player vehicle has been added/removed/changed. Update whether the player is currently in a vehicle on the {@link engUiFactory}.
         */
        $scope.$on("engIsPlayerInAVehicle", (_event, isPlayerInAVehicle) => engUiFactory.setPlayerInAVehicle( /** @type boolean */ isPlayerInAVehicle));

        /**
         * The timer has elapsed further on the server. Update the stored value on the {@link engChaosModeFactory}
         */
        $scope.$on("engChaosModeTimer", (_event, timeLeftRatio) => {
          const timer = engMiscFactory.isDefined(timeLeftRatio) ? /** @type number */ timeLeftRatio : null;
          engChaosModeFactory.setTimer(timer);
        });

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engUiFactory.view.unsubscribe(viewSubscription);
          engExplosivesFactory.configuredExplosives.unsubscribe(configuredExplosivesSubscription);
        });

        // $scope.$on("UINavigation", (_event, action, value) => {
          // console.log("UINavigation", action, value);
        // });

        init();
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engExplosiveContainer
   * @description A container component that allows configuration of an explosive by the user
   * @requires engExplosivesFactory
   * @requires engMiscFactory
   * @requires engEndpointsFactory
   * @requires engPreviewsFactory
   * @requires engUiFactory
   */
  .directive("engExplosiveContainer", ["engExplosivesFactory", "engMiscFactory", "engEndpointsFactory", "engPreviewsFactory", "engUiFactory", (
      engExplosivesFactory,
      engMiscFactory,
      engEndpointsFactory,
      engPreviewsFactory,
      engUiFactory
  ) => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/containers/explosive-container/explosive-container.html",
      restrict: "E",
      scope: false,
      controller: $scope => {
        /** @type ExplosiveDto[] */
        $scope.explosives = [];
        /** @type {ConfiguredExplosive | undefined} */
        $scope.selectedExplosive = undefined;

        /** @type ButtonControls */
        $scope.cycleExplosives = { disablePrevious: true, disableNext: true };
        /** @type ButtonControls */
        $scope.cyclePresets = { disablePrevious: true, disableNext: true };
        /** @type {{ explosive: HorizontalDirection, preset: HorizontalDirection }} */
        $scope.animationDirection = { explosive: "left", preset: "left" };

        $scope.lockTier4s = false;
        $scope.isPlayerInAVehicle = false;
        $scope.isPositioning = false;
        /** @type ConfiguredExplosive */
        $scope.explosiveShowingInformation = undefined;
        /** @type CoordinateInputType */
        $scope.selectedCoordinateOption = undefined;
        /** @type string */
        $scope.previewId = undefined;

        /**
         * A predicate that returns `true` if the given {@link ExplosiveDto} has a matching id to the
         * {@link selectedExplosive}.
         * @param {ExplosiveDto["id"]} id The id to check
         * @returns {boolean} Whether the given explosive's id matches
         */
        const equalsSelected = ({ id }) => id === $scope.selectedExplosive.id;

        /**
         * A predicate that returns `true` if the given {@link ExplosivePreset} has a matching id to the one on the
         * {@link selectedExplosive}.
         * @param {ExplosivePreset["id"]} id The id to check
         * @returns {boolean} Whether the given explosive preset's id matches
         */
        const equalsSelectedPreset = ({ id }) => id === $scope.selectedExplosive.preset.id;

        /**
         * Sets the currently selected explosive to the new given explosive
         * @param {ConfiguredExplosive} explosive The new selected explosive
         */
        const selectedExplosiveSubscription = explosive => { $scope.selectedExplosive = explosive; };

        /**
         * Sets whether tier 4 explosives are currently locked from being selected
         * @param {boolean} locked The new value
         */
        const lockTier4sSubscription = locked => { $scope.lockTier4s = locked; };

        /**
         * Sets whether the player is currently in a vehicle or not to the given value
         * @param {boolean} isPlayerInAVehicle The new value
         */
        const isPlayerInAVehicleSubscription = isPlayerInAVehicle => {
          $scope.isPlayerInAVehicle = isPlayerInAVehicle;
        };

        /**
         * Sets whether to show boss information
         * @param {boolean} showingInformation Whether to show boss information
         */
        const isShowingBossInformationSubscription = showingInformation => {
          $scope.isShowingBossInformation = showingInformation;
        };

        /**
         * Sets the preview ID to the first element in the given array
         * @param {Preview[]} previews The new previews
         */
        const previewSubscription = previews => {
          if (previews.length > 1) {
            console.warn("More than one preview is active. This is not expected.");
          }

          $scope.previewId = previews.length > 0 ? previews[0].previewId : undefined;
        };

        /**
         * Sets whether an explosive is currently being positioned on the UI to the given value
         * @param {boolean} isPositioning The new value
         */
        const isPositioningSubscription = isPositioning => { $scope.isPositioning = isPositioning; };

        /**
         * Returns a new {@link ButtonControls} object based on an array with a given length and the selected index of
         * the array the controls would be cycling through.
         * More formally, the [disablePrevious]{@link ButtonControls.disablePrevious} will be set to whether the given
         * {@link index} is equal to `0`, and the [disableNext]{@link ButtonControls.disableNext} will be set to whether
         * the given {@link index} is 1 less than the given {@link arrayLength}.
         * @param {number} index The index
         * @param {number} arrayLength The length of the array. This should be greater than `0`.
         * @returns {ButtonControls} The constructed controls
         * @example
         * newButtonControlsByArray(0, 10) => { disablePrevious: true, disableNext: false }
         * newButtonControlsByArray(5, 6) => { disablePrevious: false, disableNext: true }
         */
        function newButtonControlsByArray(index, arrayLength) {
          return { disablePrevious: index === 0, disableNext: index === arrayLength - 1 };
        }

        /**
         * Selects the next explosive in the list, relative to the currently selected explosive.
         * The newly selected explosive will default to the first preset in the explosive's preset list and will
         * maintain the currently configured {@link Position} of the {@link selectedExplosive}.
         * If {@link selectNext} is `false`, then this function selects the previous explosive in the list.
         * @param {boolean} selectNext Whether to select the next explosive (`true`), or the previous (`false`)
         */
        function updateExplosive(selectNext) {
          const index = $scope.explosives.findIndex(equalsSelected);
          const newIndex = index + (selectNext ? 1 : -1);
          const dto = $scope.explosives[newIndex];

          const newPresetIndex = 0;

          const nextExplosive = /** @type ConfiguredExplosive */ {
            ...engExplosivesFactory.dtoToExplosive(dto, newPresetIndex),
            position: $scope.selectedExplosive.position // Persist location
          };

          engExplosivesFactory.setConfiguredExplosiveByIndex(nextExplosive);
          $scope.cycleExplosives = newButtonControlsByArray(newIndex, $scope.explosives.length);
          $scope.cyclePresets = newButtonControlsByArray(newPresetIndex, dto.presets.length);
        }

        /**
         * Selects the next explosive preset in the currently selected explosive's presets list.
         * If {@link selectNext} is `false`, then this function selects the previous preset.
         * @param {boolean} selectNext Whether to select the next preset (`true`), or the previous (`false`)
         */
        function updatePreset(selectNext) {
          const { presets } = $scope.explosives.find(equalsSelected);
          const index = presets.findIndex(equalsSelectedPreset);
          const newIndex = index + (selectNext ? 1 : -1);

          engExplosivesFactory.setSelectedPresetByIndex(newIndex);
          $scope.cyclePresets = newButtonControlsByArray(newIndex, presets.length);
        }

        /**
         * Sets the {@link animationDirection} with the given list of partial updates
         * @param {{ explosive: HorizontalDirection | undefined, preset: HorizontalDirection | undefined }} animationDirection
         * A partial object to update keys inside {@link animationDirection} with
         */
        function updateAnimation(animationDirection) {
          $scope.animationDirection = { ...$scope.animationDirection, ...animationDirection };
        }

        /**
         * Sets the selected explosive's position to the given one
         * @param {Position} position The position to use
         */
        function setExplosivePosition(position) {
          const { x, y, z } = position;
          $scope.selectedExplosive.position = {
            x: engMiscFactory.roundTo(x, 3),
            y: engMiscFactory.roundTo(y, 3),
            z: engMiscFactory.roundTo(z, 3)
          };
        }

        /**
         * Creates the in-world preview by calling various related endpoints on the {@link engEndpointsFactory}:
         *  - If the given {@link option} is `"manual"`, then the {@link position} will be used to place this preview
         *  - In all other cases, the {@link option} will be sent along with the request. and the {@link position}
         *    ignored
         * @param {CoordinateInputType} option The option to use for getting the position to place the preview at
         * @param {Position} [position] The position to place the preview at. Should be given if {@link option} is
         * `"manual"`. In all other cases, this parameter is ignored even if it is given
         */
        function createPreview(option, position) {
          const { uiId, id: explosiveId } = $scope.selectedExplosive;
          const presetId = $scope.selectedExplosive.preset.id;

          /** @param {string} id The id of the preview to add */
          const addPreview = id => engPreviewsFactory.addPreview({ previewId: id, explosiveUiId: uiId });

          if (option !== "manual") {
            engEndpointsFactory.createPreviewByOption(option, explosiveId, presetId, ({ id, position: serverPosition }) => {
              addPreview(id);
              // Update from the position returned
              setExplosivePosition(serverPosition);
            });
            return;
          }

          if (!engMiscFactory.isDefined(position)) {
            console.error("position is mandatory if the input type is 'manual'");
            return;
          }

          engEndpointsFactory.createPreviewByPosition(position, explosiveId, presetId, addPreview);
        }

        /**
         * Updates the in-game preview using the currently {@link selectedCoordinateOption} and
         * {@link selectedExplosive}'s position.
         */
        function updatePreviewForCurrentlySelected() {
          if ($scope.selectedCoordinateOption === "manual") {
            const position = $scope.selectedExplosive.position;
            // If the coordinates are not fully defined, then it probably means the user has not finished typing.
            // In any case, do not send these coordinates to the server.
            if (!engMiscFactory.isDefined(position.x) || !engMiscFactory.isDefined(position.y) || !engMiscFactory.isDefined(position.z)) {
              return;
            }

            engEndpointsFactory.updatePreviewPosition($scope.previewId, position);
            return;
          }

          engEndpointsFactory.updatePreviewPositionByType(
              $scope.previewId,
              $scope.selectedCoordinateOption,
              setExplosivePosition
          );
        }

        /**
         * Sets the {@link selectedCoordinateOption} to `"manual"` if the {@link selectedExplosive}'s position is
         * defined (see ...). Otherwise, sets this option to the {@link defaultOption}.
         * If the position has been created by the player (i.e. it was persisted), then we don't want to overwrite them.
         * @param {CoordinateInputType} defaultOption
         */
        function updateCoordinateOption(defaultOption) {
          $scope.selectedCoordinateOption = engMiscFactory.isPositionDefined($scope.selectedExplosive.position)
              ? "manual"
              : defaultOption;
        }

        /**
         * Closes the positioning window and deletes the in-game preview.
         * This function will also default any `undefined` coordinate to `0`, and enforce the validation on the fuse
         * property.
         */
        function closePositioningWindow() {
          engEndpointsFactory.deletePreview($scope.previewId);
          engPreviewsFactory.setPreviews([]);
          engUiFactory.setIsPositioningExplosive(false);

          // Default any coordinate to `0` if not given when the window is closed
          const { position, fuse } = $scope.selectedExplosive;
          position.x = engMiscFactory.withDefaultValue(position.x, 0);
          position.y = engMiscFactory.withDefaultValue(position.y, 0);
          position.z = engMiscFactory.withDefaultValue(position.z, 0);

          // Enforce fuse validation
          $scope.selectedExplosive.fuse = fuse >= 0 && fuse <= 999 ? fuse : 0;
        }

        /**
         * Displays the positioning layer and shows the in-game preview, requesting it to be created if it does not
         * exist yet.
         */
        function onSelectClicked() {
          // If the animation is playing, ignore any further clicks on this button
          if ($scope.isPositioning) return;

          engUiFactory.setIsPositioningExplosive(true);
          updateCoordinateOption($scope.selectedCoordinateOption);

          if ($scope.previewId !== undefined) {
            console.warn("previewId should not be defined at this stage");
          }

          createPreview($scope.selectedCoordinateOption, $scope.selectedExplosive.position);
        }

        /**
         * Updates the selected explosive to the previous in the array of DTOs.
         * Updates the animations for both explosive and preset to `"right"`.
         */
        function onPreviousExplosiveClicked() {
          updateExplosive(false);
          // Overwrite preset animation as well to match the explosive
          updateAnimation({ explosive: "right", preset: "right" });
        }

        /**
         * Sets the selected explosive to the next in the array of DTOs.
         * Sets the animations for both explosive and preset to `"left"`.
         */
        function onNextExplosiveClicked() {
          updateExplosive(true);
          // Overwrite preset animation as well to match the explosive
          updateAnimation({ explosive: "left", preset: "left" });
        }

        /**
         * Selects the previous preset in the currently selected explosive's preset list.
         * Sets the preset animation to be `"left"`.
         */
        function onPreviousPresetClicked() {
          updatePreset(false);
          updateAnimation({ preset: "left" });
        }

        /**
         * Selects the next preset in the currently selected explosive's preset list.
         * Sets the preset animation to be `"right"`.
         */
        function onNextPresetClicked() {
          updatePreset(true);
          updateAnimation({ preset: "right" });
        }

        /**
         * Closes the positioning window
         */
        function onOverlayClicked() {
          closePositioningWindow();
        }

        /**
         * If the selected explosive is the `"convergence"` explosive, this function will set the information flag to
         * `true` on the {@link engUiFactory}.
         * Otherwise, this function hides the positioning window and preview and sets the shown view to the `"overview"`
         * page.
         */
        function onPositionPlaceClicked() {
          // This will be true if the place button has been clicked twice in quick succession.
          // We are navigating away from the position screen, but the button can still be clicked while the animation
          // is being displayed.
          if (!engMiscFactory.isDefined($scope.selectedExplosive)) return;

          if ($scope.selectedExplosive.id === "convergence") {
            engUiFactory.setIsShowingInformation(true);
            return;
          }

          closePositioningWindow();
          engUiFactory.setView("overview");
        }

        /**
         * Updates the preview for the currently {@link selectedExplosive}'s new position
         */
        function onPositionChanged() {
          updatePreviewForCurrentlySelected();
        }

        /**
         * Sets the fuse for the current explosive with the given value
         * @param {number} fuse The new fuse to set
         */
        function onFuseChanged(fuse) {
          $scope.selectedExplosive.fuse = fuse;
        }

        /**
         * Updates the {@link selectedCoordinateOption} with the given {@link option}.
         * This function will then {@link updatePreviewForCurrentlySelected} option.
         * @param {CoordinateInputType} option The new coordinate option
         */
        function onCoordinateOptionUpdated(option) {
          $scope.selectedCoordinateOption = option;
          updatePreviewForCurrentlySelected(option);
        }

        /**
         * Hides the showing of the information box on the {@link engUiFactory}.
         * This function will then place closing the position window and change to the `"overview"` view on the
         * {@link engUiFactory}. It is important to note that this will happen after a delay, which coincides with the
         * animation delay of the information box closing. This is internally using a simple timeout, and has been
         * implemented like this as the ngAnimate directive (when animating ng-if directives) does not allow hooking on
         * to animation events, from that element.
         */
        function onInformationClosed() {
          engUiFactory.setIsShowingInformation(false);

          setTimeout(() => {
            closePositioningWindow();
            engUiFactory.setView("overview");
          }, 500);
        }

        /**
         * Initialises the component.
         * Initialises the {@link engExplosivesFactory} and subscribes to all necessary subscriptions on all used
         * factories.
         *
         * This function will set the list of {@link explosives} from this factory as well as then set up
         * the {@link cycleExplosives} and {@link cyclePresets} once the list is retrieved.
         *
         * This function will additionally set the {@link selectedCoordinateOption} to `"manual"` if the selected
         * explosive has a defined position (see [isPositionDefined]{@link engMiscFactory.isPositionDefined}), otherwise
         * it is set to `"player"` if a player vehicle exists, otherwise `"camera"` as a default.
         */
        function init() {
          engPreviewsFactory.previews.subscribe(previewSubscription);
          engUiFactory.isPositioningExplosive.subscribe(isPositioningSubscription);
          engUiFactory.isPlayerInAVehicle.subscribe(isPlayerInAVehicleSubscription);
          engUiFactory.isShowingInformation.subscribe(isShowingBossInformationSubscription);

          engExplosivesFactory.init(() => {
            engExplosivesFactory.selectedExplosive.subscribe(selectedExplosiveSubscription);
            engExplosivesFactory.lockTier4s.subscribe(lockTier4sSubscription);
            $scope.explosives = engExplosivesFactory.getExplosiveDTOs();

            const index = $scope.explosives.findIndex(equalsSelected);
            const dto = $scope.explosives[index];
            const presetIndex = dto.presets.findIndex(equalsSelectedPreset);
            $scope.cycleExplosives = newButtonControlsByArray(index, $scope.explosives.length);
            $scope.cyclePresets = newButtonControlsByArray(presetIndex, dto.presets.length);

            const defaultOption = $scope.isPlayerInAVehicle ? "player" : "camera";
            updateCoordinateOption(defaultOption);

            // If the previewId already exists, don't try to create a new preview
            if ($scope.isPositioning && $scope.previewId === undefined) {
              createPreview($scope.selectedCoordinateOption, $scope.selectedExplosive.position);
            }
          });
        }

        $scope.onSelectClicked = onSelectClicked;
        $scope.onPreviousExplosiveClicked = onPreviousExplosiveClicked;
        $scope.onNextExplosiveClicked = onNextExplosiveClicked;
        $scope.onPreviousPresetClicked = onPreviousPresetClicked;
        $scope.onNextPresetClicked = onNextPresetClicked;
        $scope.onOverlayClicked = onOverlayClicked;
        $scope.onPositionPlaceClicked = onPositionPlaceClicked;
        $scope.onPositionChanged = onPositionChanged;
        $scope.onFuseChanged = onFuseChanged;
        $scope.onCoordinateOptionUpdated = onCoordinateOptionUpdated;
        $scope.onInformationClosed = onInformationClosed;

        /**
         * Unsubscribes from any subscribed observables
         */
        $scope.$on("$destroy", () => {
          engExplosivesFactory.selectedExplosive.unsubscribe(selectedExplosiveSubscription);
          engExplosivesFactory.lockTier4s.unsubscribe(lockTier4sSubscription);
          engPreviewsFactory.previews.unsubscribe(previewSubscription);
          engUiFactory.isPlayerInAVehicle.unsubscribe(isPlayerInAVehicleSubscription);
          engUiFactory.isPositioningExplosive.unsubscribe(isPositioningSubscription);
          engUiFactory.isShowingInformation.unsubscribe(isShowingBossInformationSubscription);
        });

        init();
      }
    }
  }])

angular.module("beamng.apps")
    /**
     * @ngdoc directive
     * @name beamng.apps:engChaosModeContainer
     * @description A container component that manages the chaos mode of the mod
     * @requires engUiFactory
     * @requires engChaosModeFactory
     * @requires engEndpointsFactory
     */
    .directive("engChaosModeContainer", ["engUiFactory", "engChaosModeFactory", "engEndpointsFactory", (
        engUiFactory,
        engChaosModeFactory,
        engEndpointsFactory
    ) => {
      // noinspection JSUnusedGlobalSymbols
      return {
        templateUrl: "../../modules/apps/ExplosionController/containers/chaos-mode-container/chaos-mode-container.html",
        restrict: "E",
        controller: $scope => {
          /** @type ConfiguredExplosive[] */
          $scope.explosiveOptions = [];
          /** @type ChaosModeConfiguration */
          $scope.configuration = undefined;
          $scope.isActive = false;
          $scope.selectedExplosivesErrors = [];

          /**
           * Sets the current configuration to the new given configuration.
           * Additionally, this function will enforce validation and set the selected explosives errors with any found.
           * @param {ChaosModeConfiguration} configuration The new configuration
           */
          const configurationSubscription = configuration => {
            $scope.configuration = configuration;
            $scope.selectedExplosivesErrors = configuration.allowedExplosives.length === 0
                ? ["Select at least one explosive"]
                : [];
          };

          /**
           * Sets the list of valid explosive options to the given ones.
           * @param {ConfiguredExplosive[]} explosiveOptions The new explosive options
           */
          const selectedExplosiveSubscription = explosiveOptions => { $scope.explosiveOptions = explosiveOptions; };

          /**
           * Sets whether the mode is active or not
           * @param {ConfiguredExplosive[]} isActive Whether the mode is currently active
           */
          const isActiveSubscription = isActive => { $scope.isActive = isActive; };

          /**
           * Sets the view back to the `"overview"` view on the {@link engUiFactory}
           */
          function onBackClick() {
            engUiFactory.setView("overview");
          }

          /**
           * Toggles the chaos mode on the {@link engEndpointsFactory} and the {@link engChaosModeFactory}, before
           * setting the view back to the `"overview"` view on the {@link engUiFactory}.
           */
          function onToggleClick() {
            const isNowActive = !$scope.isActive;
            engChaosModeFactory.setActive(isNowActive);

            if (!isNowActive) {
              engEndpointsFactory.stopChaosMode();
              return;
            }

            engEndpointsFactory.startChaosMode($scope.configuration);
            engUiFactory.setView("overview");
          }

          /**
           * Updates the configuration on the {@link engChaosModeFactory} with the current configuration, replacing the
           * list of [allowedExplosives]{@link ChaosModeConfiguration.allowedExplosives} with the given options.
           * @param {string[]} options The newly selected options
           */
          function onSelectedOptionsChange(options) {
            /** @type ChaosModeConfiguration */
            const newConfiguration = { ...$scope.configuration, allowedExplosives: options };
            engChaosModeFactory.setConfiguration(newConfiguration);
          }

          /**
           * Initialises the component.
           * Initialises the {@link engChaosModeFactory} and subscribes to all necessary subscriptions on all used
           * factories.
           */
          function init() {
            engChaosModeFactory.configuration.subscribe(configurationSubscription);
            engChaosModeFactory.explosiveOptions.subscribe(selectedExplosiveSubscription);
            engChaosModeFactory.isActive.subscribe(isActiveSubscription);
          }

          /**
           * Unsubscribes from any subscribed observables
           */
          $scope.$on("$destroy", () => {
            engChaosModeFactory.configuration.unsubscribe(configurationSubscription);
            engChaosModeFactory.explosiveOptions.unsubscribe(selectedExplosiveSubscription);
            engChaosModeFactory.isActive.unsubscribe(isActiveSubscription);
          });

          $scope.onBackClick = onBackClick;
          $scope.onToggleClick = onToggleClick;
          $scope.onSelectedOptionsChange = onSelectedOptionsChange;

          init();
        }
      }
    }]);

angular.module("beamng.apps")
    /**
     * @ngdoc directive
     * @name beamng.apps:engBossBarContainer
     * @description A container component that displays a multiphase boss bar, currently specifically tailored towards
     * the "The Convergence" explosive.
     * @param {boolean} showAnimations Whether to show animations on the boss bar or not
     * @requires engBossesFactory
     * @requires engMiscFactory
     */
    .directive("engBossBarContainer", ["engBossesFactory", "engMiscFactory", (engBossesFactory, engMiscFactory) => {
      // noinspection JSUnusedGlobalSymbols
      return {
        templateUrl: "../../modules/apps/ExplosionController/containers/boss-bar-container/boss-bar-container.html",
        restrict: "E",
        scope: {
          showAnimations: "<"
        },
        controller: $scope => {
          /** @type ConvergencePhaseProgress */
          $scope.mainPhase = undefined;
          /** @type ConvergencePhaseProgress[] */
          $scope.phases = [];
          $scope.failedSegment = false;
          $scope.failedMainPhaseSegment = false;

          /**
           * Sets the phases list and main phases to the given progress. If the progress is `null`, then the phases list
           * is set to an empty array and the main phase is set to `undefined`.
           * @param {ConvergencePhaseProgress[] | null} progress The next progress list
           */
          const bossBattleProgressSubscription = progress => {
            if (!engMiscFactory.isDefined(progress)) {
              $scope.phases = [];
              $scope.mainPhase = undefined;
              return;
            }

            const lastElementIndex = progress.length - 1;
            $scope.phases = progress.slice(0, lastElementIndex);
            $scope.mainPhase = progress[lastElementIndex];
          }

          /**
           * Sets the current segments has been failed to the given flag.
           * This function will also set whether the main phase segment was failed, which will become `true` if all over
           * phases are dead and the given flag is itself `true`.
           * @param {boolean} failed The next failed flag value
           */
          const setHasFailedSegmentSubscription = failed => {
            $scope.failedSegment = failed;
            $scope.failedMainPhaseSegment = $scope.phases.every(({ dead }) => dead) && failed;
          };

          /**
           * Updates the {@link engBossesFactory} with the new progress data
           * @param {ConvergencePhaseProgress[] | undefined} data The progress data
           */
          function onUpdate(data) {
            const newData = data !== undefined ? data : null;
            engBossesFactory.setBossBattleProgress(newData);
          }

          /**
           * Updates whether the player has failed the current segment of the boss on the {@link engBossesFactory}
           * @param {boolean} failed Whether the player has failed the current segment of the boss
           */
          function onFailedUpdate(failed) {
            engBossesFactory.setHasFailedSegment(failed);
          }

          /**
           * Initialises the component.
           * Subscribes to various subscriptions on the {@link engBossesFactory}.
           */
          function init() {
            engBossesFactory.bossBattleProgress.subscribe(bossBattleProgressSubscription);
            engBossesFactory.hasFailedSegment.subscribe(setHasFailedSegmentSubscription);
          }

          $scope.$on("engBossBarUpdate", (_event, data) => onUpdate(/** @type {ConvergencePhaseProgress[] | undefined} */ data));

          $scope.$on("engBossBarFailed", (_event, data) => onFailedUpdate(/** @type boolean */ data));

          /**
           * Unsubscribes from any subscribed observables
           */
          $scope.$on("$destroy", () => {
            engBossesFactory.bossBattleProgress.unsubscribe(bossBattleProgressSubscription);
            engBossesFactory.hasFailedSegment.unsubscribe(setHasFailedSegmentSubscription);
          });

          init();

          // setTimeout(() => {
          //   onUpdate([
          //     { active: true, dead: false, segments: 3, segmentsCompleted: 0 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 2 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 0 },
          //     { active: true, dead: false, segments: 6, segmentsCompleted: 3 }
          //   ]);
          //   onFailedUpdate(true);
          // }, 1000);
          //
          // setTimeout(() => {
          //   onUpdate([
          //     { active: true, dead: true, segments: 3, segmentsCompleted: 3 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 2 },
          //     { active: false, dead: false, segments: 3, segmentsCompleted: 0 },
          //     { active: true, dead: false, segments: 6, segmentsCompleted: 3 }
          //   ]);
          //   onFailedUpdate(true);
          // }, 2500);
        }
      }
    }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engSlider
   * @description An input slider that allows selecting a single value. along a linear axis
   * @param {string} label The label to display on the slider
   * @param {number} max The maximum value
   * @param {number} model The current value of the slider. This is updated in-place.
   * @param {number} [min] The minimum value. Defaults to `0`.
   * @param {boolean} [disabled] Whether the slider is disabled and cannot be selected
   */
  .directive("engSlider", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/slider/slider.html",
      restrict: "E",
      transclude: true,
      scope: {
        label: "<",
        max: "<",
        model: "=",
        min: "<",
        disabled: "<"
      },
      link: $scope => {
        $scope.disabled = $scope.disabled || false;
        $scope.min = $scope.min || 0;
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engSlide
   * @description A wrapper to an element that animates sliding in and out from a direction. The `direction` property
   * specifies the direction of the animation sliding in; the animation will slide out in the opposite direction.
   * @param {boolean} enter Whether the contents are visible and in-frame
   * @param {Direction} direction The direction the animation will enter
   * @param {boolean} [fullSize] Whether the slide animation should fit to the available space (both horizontally and
   * vertically). Optional and defaults to `false`.
   * @param {boolean} [animate] Whether the slide animation should be played at all. Defaults to `true`.
   */
  .directive("engSlide", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/slide/slide.html",
      restrict: "E",
      transclude: true,
      scope: {
        enter: "<",
        direction: "<",
        fullSize: "<",
        animate: "<"
      },
      link: ($scope) => {
        $scope.fullSize = $scope.fullSize || false;
        $scope.animate = $scope.animate || true;
      }
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engProgressBar
   * @description Displays a value between (0, 1) as a progress bar in a specified colour, with a text label
   * @param {string} label The text to display as a label above the progress bar
   * @param {number} value The value  between (0, 1) to display
   * @param {boolean} [glow] Whether the value should glow
   */
  .directive("engProgressBar", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/progress-bar/progress-bar.html",
      restrict: "E",
      scope: {
        label: "<",
        value: "<",
        glow: "<"
      },
      link: $scope => {
        $scope.glow = $scope.glow || false;
      }
    };
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engPositioning
   * @description Provides an interface for a user to select how to configure a 3D coordinate position with an
   * additional fuse. This uses 2-way binding on the inputted {@link model}, meaning that this value will be updated in
   * place. On any coordinate change, this directive will emit on the {@link positionChange} event.
   * @param {{ position: Position, fuse: number }} model The model to configure. This will be modified in place.
   * @param {CoordinateInputType} selectedOption The selected option for providing coordinates
   * @param {boolean} [allowVehiclePositioning] Whether to allow selecting by vehicle position. Optional and defaults to false.
   * @param {() => void} placeClick A callback function that will be called when the `place` button is clicked
   * @param {() => void} positionChange A callback function that will be called when any of the coordinates have been modified
   * @param {(option: CoordinateInputType) => void} coordinateOptionUpdate A callback function that will be called when
   * the option/method of input has been changed.
   */
  .directive("engPositioning", [() => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/positioning/positioning.html",
      restrict: "E",
      scope: {
        model: "=",
        selectedOption: "<",
        placeClick: "<",
        allowVehiclePositioning: "<",
        positionChange: "<",
        coordinateOptionUpdate: "<"
      },
      link: $scope => {
        $scope.allowVehiclePositioning = $scope.allowVehiclePositioning || false;

        /**
         * Emits the currently selected option through the update function
         */
        function onRefreshClick() {
          $scope.coordinateOptionUpdate($scope.selectedOption);
        }

        /**
         * If the key that was pressed was the `"Enter"` key, then this function will emit the `placeClick` function
         * @param {KeyboardEvent} event The keyboard event
         */
        function onInputKeyPressed(event) {
          if (event.key === "Enter") {
            $scope.placeClick();
          }
        }

        $scope.onRefreshClick = onRefreshClick;
        $scope.onInputKeyPressed = onInputKeyPressed;
      }
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engPanel
   * @description Provides a container with a header which additionally has main and footer content elements projected
   * into it. Use the `<engPanelContent>` and `<engPanelFooter>` to inject main and footer content respectively.
   * The `<engPanelActionBar>` is optional and should be used for action bars below the title.
   * @param {string} title The title to put in the header of the panel
   */
  .directive("engPanel", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/panel/panel.html",
      restrict: "E",
      transclude: {
        actionBar: "?engPanelActionBar",
        content: "engPanelContent",
        footer: "engPanelFooter"
      },
      scope: {
        title: "<"
      }
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engMaterialsWarning
   * @description A dialogue that warns the player about missing materials for the current level
   */
  .directive("engMaterialsWarning", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/materials-warning/materials-warning.html",
      restrict: "E",
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engIconButton
   * @description A button which contains an icon, rather than text
   * @param {string} icon The name of the Material icon to use
   * @param {boolean} [disabled] Whether the button is disabled and cannot be clicked. Optional and defaults to `false`.
   * @param {IconButtonSize} [size] The size of the button. Defaults to `"medium"`.
   * @param {() => void} onClick A callback function to call when the button is clicked
   */
  .directive("engIconButton", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/icon-button/icon-button.html",
      restrict: "E",
      transclude: true,
      scope: {
        icon: "<",
        size: "<",
        disabled: "<",
        onClick: "<"
      },
      link: $scope => {
        $scope.disabled = $scope.disabled || false;
        $scope.size = $scope.size || "medium";
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engExplosiveProperty
   * @description Displays an {@link ExplosiveProperty}
   * @param {ExplosiveProperty} property The property to display
   */
  .directive("engExplosiveProperty", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/explosive-property/explosive-property.html",
      restrict: "E",
      scope: {
        property: "<"
      }
    };
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engExplosiveOverview
   * @description Displays an overview of an {@link ConfiguredExplosive}
   * @param {ConfiguredExplosive | null} explosive The configured explosive to show an overview of. This can be `null`
   * @param {() => void} openClick A callback function to call when the explosive is clicked
   * @param {() => void} [deleteClick] A callback function to call when the `delete` button is clicked. This is not
   * required (and will be ignored) if the {@link explosive} is `null`.
   * @param {() => void} [previewClick] A callback function to call when the `preview` button is clicked. This is not
   * required (and will be ignored) if the {@link explosive} is `null`.
   */
  .directive("engExplosiveOverview", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/explosive-overview/explosive-overview.html",
      restrict: "E",
      scope: {
        explosive: "<",
        openClick: "&",
        deleteClick: "&",
        previewClick: "&"
      },
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engExplosive
   * @description Displays a detailed view of a {@link ConfiguredExplosive} object
   * @param {ConfiguredExplosive} explosive The explosive to display
   * @param {ButtonControls} cycleExplosives A set of boolean values for controlling whether the previous and next
   * buttons for the explosive banner are disabled and cannot be clicked by the user
   * @param {ButtonControls} cyclePresets A set of boolean values for controlling whether the previous and next buttons
   * for the preset banner are disabled and cannot be clicked by the user
   * @param {HorizontalDirection} explosiveAnimationDirection The direction for which the explosive banner's text will
   * animate from
   * @param {HorizontalDirection} presetAnimationDirection The direction for which the explosive's preset banner's text
   * will animate from
   * @param {boolean} [locked] Whether the explosive is locked and cannot be selected
   * @param {() => void} nextExplosiveClick A callback function to call when the `previous` explosive arrow is clicked
   * @param {() => void} previousExplosiveClick A callback function to call when the `next` explosive arrow is clicked
   * @param {() => void} nextPresetClick A callback function to call when the `previous` preset arrow is clicked
   * @param {() => void} previousPresetClick A callback function to call when the `next` preset arrow is clicked
   * @param {() => void} selectClick A callback function to call when the `select` button is clicked
   * @param {() => void} backClick A callback function to call when the `back` button is clicked
   */
  .directive("engExplosive", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/explosive/explosive.html",
      restrict: "E",
      scope: {
        explosive: "<",
        cycleExplosives: "<",
        cyclePresets: "<",
        explosiveAnimationDirection: "<",
        presetAnimationDirection: "<",
        locked: "<",
        nextExplosiveClick: "<",
        previousExplosiveClick: "<",
        nextPresetClick: "<",
        previousPresetClick: "<",
        selectClick: "<"
      }
    }
  }])

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engDropdown
   * @description A dropdown list that allows selecting multiple options from a list
   * @param {string} label A label to display
   * @param {Option[]} model A list of currently selected options
   * @param {Option[]} options A list of all options
   * @param {boolean} [invalid] Whether the dropdown state is invalid. Defaults to `false`.
   * @param {string[]} [errors] A list of errors to show. Defaults to an empty list.
   * @param {string} [placeholder] An optional placeholder to show when no items are selected
   * @param {boolean} [disabled] Whether the dropdown is disabled. Defaults to `false`.
   * @param {(options: Option[]) => void} onChange A callback function to call when the list of options is updated. The
   * new list is given in the callback and will be called _before_ the changes are reflected in the {@link model}.
   */
  .directive("engDropdown", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/dropdown/dropdown.html",
      restrict: "E",
      transclude: true,
      scope: {
        label: "<",
        model: "=",
        options: "<",
        invalid: "<",
        errors: "<",
        placeholder: "<",
        disabled: "<",
        onChange: "<"
      },
      link: $scope => {
        $scope.invalid = $scope.invalid || false;
        $scope.errors = $scope.errors || [];
        $scope.disabled = $scope.disabled || false;
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engDialogue
   * @description Displays an inline text box containing an icon and projected text, that comes in multiple variants to
   * indicate severity/importance.
   * @param {DialogueVariant} variant The variant of the dialogue. Should be used to indicate severity/importance.
   * @param {string} icon The icon to use
   */
  .directive("engDialogue", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/dialogue/dialogue.html",
      restrict: "E",
      transclude: true,
      scope: {
        variant: "<",
        icon: "<"
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engButton
   * @description A button which provides styling for a few variants within the mod
   * @param {string} text The text to display on the button
   * @param {ButtonVariant} [variant] The variant of the button to use. Optional and defaults to `"primary"`.
   * @param {boolean} [disabled] Whether the button is disabled and cannot be clicked. Optional and defaults to `false`.
   * @param {() => void} onClick A callback function to call when the button is clicked
   */
  .directive("engButton", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/button/button.html",
      restrict: "E",
      transclude: true,
      scope: {
        text: "<",
        variant: "<",
        disabled: "<",
        onClick: "<"
      },
      link: $scope => {
        $scope.variant = $scope.variant || "primary";
        $scope.disabled = $scope.disabled || false;
      }
    }
  }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engBossInformation
   * @description Displays boss information, currently specific to the "The Convergence" explosive
   * @param {Difficulty} difficulty The difficulty level of the boss that is being displayed
   * @param {() => void} closeClick A function to call when the `close` button is clicked
   */
  .directive("engBossInformation", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/boss-information/boss-information.html",
      restrict: "E",
      transclude: true,
      scope: {
        difficulty: "<",
        closeClick: "<"
      }
    }
  }])

angular.module("beamng.apps")
    /**
     * @ngdoc directive
     * @name beamng.apps:engBossBar
     * @description Displays an animated boss bar for an anomaly entity
     * @param {string} entity The name of the entity that this boss bar is representing
     * @param {number} segments The number of segments of life to display
     * @param {number} segmentsCompleted Te number of segments of life completed
     * @param {boolean} active Whether the entity this boss bar is representing is active
     * @param {boolean} [failedLatest] Whether the current active segment was failed
     * @param {"red" | "orange"} [colour] The colour theme to use. Defaults to `"red"`.
     * @param {boolean} [small] Whether to use a smaller height and smaller text size. Defaults to `false`.
     * @param {boolean} [dead] Whether the entity this boss bar is representing has been killed
     */
    .directive("engBossBar", [() => {
      return {
        templateUrl: "../../modules/apps/ExplosionController/components/boss-bar/boss-bar.html",
        restrict: "E",
        scope: {
          entity: "<",
          segments: "<",
          segmentsCompleted: "<",
          active: "<",
          failedLatest: "<",
          colour: "<",
          small: "<",
          dead: "<",
        },
        link: $scope => {
          $scope.colour = $scope.colour || "red";
          $scope.small = $scope.small || false;
          $scope.dead = $scope.dead || false;
          $scope.failedLatest = $scope.failedLatest || false;

          /**
           * Creates an array with the given length
           * @param length The length of the array
           * @returns {any[]} The created array
           */
          $scope.range = length => new Array(length);
        }
      }
    }]);

angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engBanner
   * @description Displays a banner that fills the width it is given. The banner provides two arrow buttons, either side
   * of an optional transcluded element, that will fire the `previousClick` and `nextClick` functions respectively.
   * @param {string} text The text to display on the banner
   * @param {ButtonControls} buttons A set of boolean values for controlling whether the previous and next buttons are
   * disabled and cannot be clicked by the user
   * @param {() => void} previousClick A callback function to call when the `previous` arrow is clicked
   * @param {() => void} nextClick A callback function to call when the `next` arrow is clicked
   * @param {boolean} [primary] Whether the banner should use the primary styling for the text. Defaults to `false`.
   * @param {HorizontalDirection} [animationDirection] The direction to animate the text when it changes. Defaults to `"left"`.
   * @param {boolean} [bottomBorder] Whether to add a bottom border to the banner. Defaults to `false`.
   */
  .directive("engBanner", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/banner/banner.html",
      restrict: "E",
      transclude: true,
      scope: {
        text: "<",
        buttons: "<",
        previousClick: "<",
        nextClick: "<",
        primary: "<",
        animationDirection: "<",
        bottomBorder: "<"
      },
      link: $scope => {
        $scope.bottomBorder = $scope.bottomBorder || false;
        $scope.animationDirection = $scope.animationDirection || "left";
        $scope.primary = $scope.primary || false;
      },
    }
  }]);


