angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engBanner
   * @description Displays a banner that fills the width it is given. The banner provides two arrow buttons, either side
   * of an optional transcluded element, that will fire the `previousClick` and `nextClick` functions respectively.
   * @param {string} text The text to display on the banner
   * @param {ButtonControls} buttons A set of boolean values for controlling whether the previous and next buttons are
   * disabled and cannot be clicked by the user
   * @param {() => void} previousClick A callback function to call when the `previous` arrow is clicked
   * @param {() => void} nextClick A callback function to call when the `next` arrow is clicked
   * @param {boolean} [primary] Whether the banner should use the primary styling for the text. Defaults to `false`.
   * @param {HorizontalDirection} [animationDirection] The direction to animate the text when it changes. Defaults to `"left"`.
   * @param {boolean} [bottomBorder] Whether to add a bottom border to the banner. Defaults to `false`.
   */
  .directive("engBanner", [() => {
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/banner/banner.html",
      restrict: "E",
      transclude: true,
      scope: {
        text: "<",
        buttons: "<",
        previousClick: "<",
        nextClick: "<",
        primary: "<",
        animationDirection: "<",
        bottomBorder: "<"
      },
      link: $scope => {
        $scope.bottomBorder = $scope.bottomBorder || false;
        $scope.animationDirection = $scope.animationDirection || "left";
        $scope.primary = $scope.primary || false;
      },
    }
  }]);
