angular.module("beamng.apps")
    /**
     * @ngdoc directive
     * @name beamng.apps:engBossBar
     * @description Displays an animated boss bar for an anomaly entity
     * @param {string} entity The name of the entity that this boss bar is representing
     * @param {number} segments The number of segments of life to display
     * @param {number} segmentsCompleted Te number of segments of life completed
     * @param {boolean} active Whether the entity this boss bar is representing is active
     * @param {boolean} [failedLatest] Whether the current active segment was failed
     * @param {"red" | "orange"} [colour] The colour theme to use. Defaults to `"red"`.
     * @param {boolean} [small] Whether to use a smaller height and smaller text size. Defaults to `false`.
     * @param {boolean} [dead] Whether the entity this boss bar is representing has been killed
     */
    .directive("engBossBar", [() => {
      return {
        templateUrl: "../../modules/apps/ExplosionController/components/boss-bar/boss-bar.html",
        restrict: "E",
        scope: {
          entity: "<",
          segments: "<",
          segmentsCompleted: "<",
          active: "<",
          failedLatest: "<",
          colour: "<",
          small: "<",
          dead: "<",
        },
        link: $scope => {
          $scope.colour = $scope.colour || "red";
          $scope.small = $scope.small || false;
          $scope.dead = $scope.dead || false;
          $scope.failedLatest = $scope.failedLatest || false;

          /**
           * Creates an array with the given length
           * @param length The length of the array
           * @returns {any[]} The created array
           */
          $scope.range = length => new Array(length);
        }
      }
    }]);