angular.module("beamng.apps")
  /**
   * @ngdoc directive
   * @name beamng.apps:engPositioning
   * @description Provides an interface for a user to select how to configure a 3D coordinate position with an
   * additional fuse. This uses 2-way binding on the inputted {@link model}, meaning that this value will be updated in
   * place. On any coordinate change, this directive will emit on the {@link positionChange} event.
   * @param {{ position: Position, fuse: number }} model The model to configure. This will be modified in place.
   * @param {CoordinateInputType} selectedOption The selected option for providing coordinates
   * @param {boolean} [allowVehiclePositioning] Whether to allow selecting by vehicle position. Optional and defaults to false.
   * @param {() => void} placeClick A callback function that will be called when the `place` button is clicked
   * @param {() => void} positionChange A callback function that will be called when any of the coordinates have been modified
   * @param {(option: CoordinateInputType) => void} coordinateOptionUpdate A callback function that will be called when
   * the option/method of input has been changed.
   */
  .directive("engPositioning", [() => {
    // noinspection JSUnusedGlobalSymbols
    return {
      templateUrl: "../../modules/apps/ExplosionController/components/positioning/positioning.html",
      restrict: "E",
      scope: {
        model: "=",
        selectedOption: "<",
        placeClick: "<",
        allowVehiclePositioning: "<",
        positionChange: "<",
        coordinateOptionUpdate: "<"
      },
      link: $scope => {
        $scope.allowVehiclePositioning = $scope.allowVehiclePositioning || false;

        /**
         * Emits the currently selected option through the update function
         */
        function onRefreshClick() {
          $scope.coordinateOptionUpdate($scope.selectedOption);
        }

        /**
         * If the key that was pressed was the `"Enter"` key, then this function will emit the `placeClick` function
         * @param {KeyboardEvent} event The keyboard event
         */
        function onInputKeyPressed(event) {
          if (event.key === "Enter") {
            $scope.placeClick();
          }
        }

        $scope.onRefreshClick = onRefreshClick;
        $scope.onInputKeyPressed = onInputKeyPressed;
      }
    }
  }])