angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engBossesFactory
   * @description A store that manages the state of active bosses. This is currently specifically tailored towards the
   * "The Convergence" explosive.
   * @requies engObservablesFactory
   */
  .factory("engBossesFactory", ["engObservablesFactory", engObservablesFactory => {
    /** @type {Observable<ConvergencePhaseProgress[] | null>} */
    const bossBattleProgress = engObservablesFactory.newObservable(null);
    /** @type {Observable<boolean>} */
    const hasFailedSegment = engObservablesFactory.newObservable(false);

    /**
     * Updates the current boss battle progress with the given progress list
     * @param {ConvergencePhaseProgress[] | null} newProgress The new progress of the boss
     */
    function setBossBattleProgress(newProgress) {
      bossBattleProgress.next(newProgress);
    }

    /**
     * Updates whether the player has failed the current segment of the boss
     * @param {boolean} failed Whether the player has failed the current segment of the boss
     */
    function setHasFailedSegment(failed) {
      hasFailedSegment.next(failed);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      bossBattleProgress.next(null);
      hasFailedSegment.next(false);
    }

    return {
      bossBattleProgress: /** @type {Subscribable<ConfiguredExplosive[] | null>} */ bossBattleProgress,
      hasFailedSegment: /** @type {Subscribable<boolean>} */ hasFailedSegment,
      setBossBattleProgress,
      setHasFailedSegment,
      reset
    };
  }]);