/**
 * @jest-environment jsdom
 */

require("angularjs-jest");
angular.module("beamng.apps", []);
require("./observables");

describe("Observable", () => {
  let observablesFactory;

  beforeEach(() => {
    const $injector = angular.injector(["beamng.apps"]);
    observablesFactory = $injector.get("engObservablesFactory");
  });

  it.each([0, "hello", null, undefined, { id: "string" }])("should set initial value to %p", (initValue) => {
    const observable = observablesFactory.newObservable(initValue);
    expect(observable.value).toEqual(initValue);
  });

  describe("subscribe", () => {
    it("should call any callbacks added to it and when the value changes", () => {
      const observable = observablesFactory.newObservable(1);
      let callbacksCalled = 0;

      observable.subscribe(() => callbacksCalled += 1);
      observable.subscribe(() => callbacksCalled += 1);
      observable.subscribe(() => callbacksCalled += 1);

      // Called initially when the subscriptions are set up
      expect(callbacksCalled).toBe(3);
      expect(observable.subscriptions.length).toBe(3);

      observable.next(10);
      expect(callbacksCalled).toBe(6);
    });
  });

  describe("unsubscribe", () => {
    it("should remove any callbacks added to it", () => {
      const observable = observablesFactory.newObservable(1);
      let callbacksCalled = 0;

      const subscription = () => callbacksCalled += 1;

      observable.subscribe(subscription);

      // Called initially when the subscriptions are set up
      expect(callbacksCalled).toBe(1);
      expect(observable.subscriptions.length).toBe(1);

      observable.unsubscribe(subscription);
      observable.next(10);

      // Shouldn't be called
      expect(callbacksCalled).toBe(1);
      expect(observable.subscriptions.length).toBe(0);
    });
  });
});
