angular.module("beamng.apps")
  /**
   * @ngdoc factory
   * @name beamng.apps:engPreviewsFactory
   * @description A store that manages the state of in-game previews that are visible
   * @requires engObservablesFactory
   */
  .factory("engPreviewsFactory", ["engObservablesFactory", (engObservablesFactory) => {
    /** @type Observable<Preview[]> */
    const previews = engObservablesFactory.newObservable([]);

    /**
     * Adds the given preview to the list
     * @param {Preview} preview The preview to add
     */
    function addPreview(preview) {
      const newPreviews = [...previews.value, { ...preview }];
      previews.next(newPreviews);
    }

    /**
     * Removes the given preview to the list
     * @param {Preview} preview The preview to remove
     */
    function removePreview(preview) {
      const newPreviews = previews.value.filter(p => p !== preview);
      previews.next(newPreviews);
    }

    /**
     * Sets the list of previews to the given list
     * @param {Preview[]} newPreviews The new previews to set
     */
    function setPreviews(newPreviews) {
      previews.next([...newPreviews]);
    }

    /**
     * Resets all data stored in this factory to the initial state
     */
    function reset() {
      previews.next([]);
    }

    return {
      previews: /** @type Subscribable<Preview[]> */ previews,
      addPreview,
      removePreview,
      setPreviews,
      reset
    };
  }]);