angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngExplosiveRankTitle
   * @description Converts an integer into its corresponding explosive rank title. Supports values between
   * (1, 10).
   */
  .filter("EngExplosiveRankTitle", () => {
    /**
     * Converts the given integer into its corresponding rank title.
     * Supports value between (1, 10).
     * @param {number} input The integer to convert
     * @returns {string} The rank description of {@link input}
     * @throws Error if the {@link input} is not an integer between (1, 10)
     */
    function getText(input) {
      switch (input) {
        case  1: return "Unsafe";
        case  2: return "Harmful";
        case  3: return "Hazardous";
        case  4: return "Destructive";
        case  5: return "Dangerous";
        case  6: return "Deadly";
        case  7: return "Highly Deadly";
        case  8: return "Extremely Deadly";
        case  9: return "Apocalyptic";
        case 10: return "Annihilistic";
        default: throw new Error(`${input} is not supported by "EngExplosiveRankTitle"`);
      }
    }

    return input => getText(input).toUpperCase();
  });