angular.module("beamng.apps")
  /**
   * @ngdoc filter
   * @name beamng.apps:EngExplosiveTierDescription
   * @description Converts an integer into its corresponding explosive tier description. Supports values between (1, 4).
   */
  .filter("EngExplosiveTierDescription", () => {
    /**
     * Converts the given integer into its corresponding tier description.
     * Supports value between (1, 4).
     * @param {number} input The integer to convert
     * @returns {string} The rank description of {@link input}
     * @throws Error if the {@link input} is not an integer between (1, 4)
     */
    function transform(input) {
      switch (input) {
        case 1: return "Tier 1 explosives are just about a legal amount of explosive power. These explosives are small and more of an annoyance than damaging.";
        case 2: return "Tier 2 explosives are where the real vehicular carnage begins: more power, more range and more difficulty in trying to survive.";
        case 3: return "Tier 3 explosives are beyond dangerous. Range becomes measured in kilometres, power becomes measured in meganewtons. Even the fabric of space and time are not safe from these explosives.";
        case 4: return "Tier 4 explosives are not just apocalyptic. Explosives this powerful will annihilate anything in their path. No vehicle is too far away from them, no vehicle is strong enough to withstand them, no vehicle is fast enough to flee from them - survival is an infinitesimally small possibility.";
        default: throw new Error(`${input} is not supported by "EngExplosiveTierDescription"`);
      }
    }

    return transform;
  });