/**
 * @external bngApi
 */
/**
 * @typedef {{
 *   engineLua: (command: string, callback?: any) => void,
 *   serializeToLua: (object: Object) => string
 * }} bngApi
 */

/**
 * A preset for a particular explosive containing statistics and details
 * @typedef {{
 *   id: string,
 *   name: string,
 *   range: number,
 *   energy: number,
 *   accuracy: number,
 *   attackRate: number,
 *   rank: number
 * }} ExplosivePreset
 */

/**
 * A property of a given explosive with a user-facing description
 * @typedef {{ id: string, name: string, description: string }} ExplosiveProperty
 */

/**
 * An explosive returned from the server with all possible presets
 * @typedef {{
 *  id: string,
 *  name: string,
 *  tier: number,
 *  description: string,
 *  properties: ExplosiveProperty[],
 *  presets: ExplosivePreset[]
 * }} ExplosiveDto
 */

/**
 * An explosive with id, name, tier, description, fuse, and properties from {@link ExplosiveDto}, with a set
 * [preset]{@link ExplosivePreset} for the explosive. Additionally, the configured explosive has:
 *  - A {@link Position}
 *  - An additional uiId for use within the client/ui and server interactions
 *  - A boolean flag for whether it is being shown
 *  - A boolean flag for whether it is being detonated
 *  - A delay between the explosive activating and it then detonating
 * @typedef {{
 *  uiId: string,
 *  id: string,
 *  name: string,
 *  tier: number,
 *  description: string,
 *  fuse: number,
 *  properties: ExplosiveProperty[],
 *  preset: ExplosivePreset,
 *  position: Position,
 *  showing: boolean,
 *  inProgress: boolean
 * }} ConfiguredExplosive
 */

/**
 * A set of properties for a given phase of the "The Convergence" explosive
 * @typedef {{ segments: number, segmentsCompleted: number, dead: boolean, active: boolean }} ConvergencePhaseProgress
 */

/**
 * A set of properties for a boss entity; currently specific for the "The Convergence" explosive
 * @typedef {{ progress: ConvergencePhaseProgress[], failed: boolean }} BossData
 */

/**
 * A subset of {@link CoordinateInputType} with just the automatic inputs (i.e. do not require a position to be known by
 * the user)
 * @typedef {"player" | "camera"} AutoCoordinateInputType
 */

/**
 * A list of all types of inputs that can be used to provide coordinates
 * @typedef {AutoCoordinateInputType | "manual"} CoordinateInputType
 */

/**
 * A set of boolean values which store whether the previous or next pair of buttons are disabled or enabled
 * @typedef {{ disablePrevious: boolean, disableNext: boolean }} ButtonControls
 */

/**
 * Directions on the horizontal axis
 * @typedef {"left" | "right"} HorizontalDirection
 */

/**
 * Directions on the horizontal and vertical axis
 * @typedef {HorizontalDirection | "top" | "bottom"} Direction
 */

/**
 * A list of all views
 * @typedef {"explosive" | "overview" | "chaos"} View
 */

/**
 * A position in 3D coordinate space
 * @typedef {{ x: number, y: number, z: number }} Position
 */

/**
 * The list of variants for the {@link engButton} component
 * @typedef {"primary" | "blinking"} ButtonVariant
 */

/**
 * The list of sizes for the {@link engIconButton} component
 * @typedef {"small" | "medium"} IconButtonSize
 */

/**
 * The list of variants for the {@link engDialogue} component
 * @typedef {"info" | "important"} DialogueVariant
 */

/**
 * An option object that has an internal id, and a displayable name
 * @typedef {{ id: string, name: string }} Option
 */

/**
 * The list of difficulties for any entity that can have a scaling difficulty
 * @typedef {"easy" | "normal" | "hard"} Difficulty
 */

/**
 * A request object to be sent to the server, from the UI, to place explosives
 * @typedef {{ explosiveId: string, presetId: string, uiId: string, position: Position, fuse: number }} PlaceExplosiveRequest
 */

/**
 * The list of identifiers that form a composite id for each configured explosive between the UI and server
 * @typedef {{ uiId: string, explosiveId: string, presetId: string, fuse: number }} ExplosiveIdentifier
 */

/**
 * A pair of ids containing the in-game preview id (from the server) and the explosive id (from the ui)
 * @typedef {{ previewId: string, explosiveUiId: string }} Preview
 */

/**
 * A configuration object for the chaos mode on the server
 * @typedef {{
 *   interval: number,
 *   maximumConcurrentExplosives: number,
 *   allowedExplosives: string[],
 *   placementRandomness: number,
 *   timer: number
 * }} ChaosModeConfiguration
 */

/**
 * A function which gets passed a value to be consumed
 * @template T
 * @typedef {(value: T) => void} SubscriptionFn
 */

/**
 * @template T
 * @typedef {{
 *   subscribe: (subscription: SubscriptionFn<T>) => void,
 *   unsubscribe: (subscription: SubscriptionFn<T>) => void
 * }} Subscribable<T>
 */

/**
 * An object that is returned from the server that holds a snapshot of the current server-side state
 * @typedef {{
 *   explosives: ExplosiveIdentifier[],
 *   bossData: BossData | null,
 *   isPlayerInAVehicle: boolean,
 *   needsReset: boolean,
 *   areMaterialsSetUp: boolean,
 *   chaosMode: ChaosModeConfiguration
 * }} StateResponse
 */